from __future__ import annotations

from enum import StrEnum
from typing import Annotated, ClassVar, Literal

from pydantic import BaseModel, Field

from .common import (
    ArtifactParam,
    ParametersBase,
    ResultBase,
    TableResult,
    TableSourceQuery,
    TableSourceTable,
    TableSourceTaskResult,
)


class OptimizationDirection(StrEnum):
    MAXIMIZE = "MAXIMIZE"
    MINIMIZE = "MINIMIZE"


class ScaleType(StrEnum):
    MULTIPLICATIVE = "MULTIPLICATIVE"
    ADDITIVE = "ADDITIVE"
    RANK = "RANK"


class AllowedMutation(BaseModel):
    in_range: RangeAnnotation = Field(
        description="The annotation expression defining the positions at which the amino acids listed in `amino_acids` can be used."
    )
    amino_acids: list[Annotated[str, Field(description="A single amino acid.")]]


class AnalyzeDataParameters(ParametersBase):
    task_type: ClassVar[str] = "analyze.data/v2"

    dataset: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(description="The data to analyze before training.")
    reference: MonomerAssembly | VhVlAssembly = Field(
        description="The reference assembly used for computing mutations."
    )
    assays: list[Assay] = Field(description="The assays present in the dataset.")


class AnalyzeDataResult(ResultBase):
    report: ArtifactParam = Field(description="The analysis report.")


class Assay(BaseModel):
    assay_id: str = Field(description="ID of the assay")
    name: str = Field(description="Human readable name of the assay")
    scale_type: ScaleType = Field(
        description="The scale type defines how the assay behaves.\n\n* **Additive**: Assay values are comparable in magnitude over batches, a delta of eg 5 in one batch is equivalent to 5 in another batch.\n* **Multiplicative**: Assay values are comparable over batches by applying a multiplication factor, eg fold improvement: `score_variantB_batch1 = factor(typically starting sequence score) * score_variantB_batch`.\n* **Rank**: Assay values are not comparable over batches, we can only assume the ranking within a batch is correct."
    )
    unit: str | None = Field(default=None, description="Unit of the assay - used for display purposes only")


class AssemblyFeature(BaseModel):
    ranges: dict[str, list[tuple[int, int]]] = Field(description="The ranges of the feature, keyed by polymer name")
    metadata: dict[str, str | int | float | bool] | None = Field(
        default=None, description="Optional metadata for the feature"
    )


class CombinatorialGenerator(BaseModel):
    type: Literal["COMBINATORIAL"] = Field(default="COMBINATORIAL")
    template: MonomerAssembly | VhVlAssembly = Field(
        description="The assembly that serves as template for generation, including any annotations to be used in `requirements`."
    )
    max_mutations: int = Field(
        description="Maximum number of mutations to combine in generation. Lower values will result in fewer total candidate sequences evaluated for this generator."
    )
    requirements: list[Requirement] = Field(
        default_factory=list,
        description="A list of requirements to satisfy. All requirements must be met for a valid assembly. See the `Requirement` documentation for details on how requirements are evaluated.",
    )
    allowed_mutations: list[AllowedMutation] = Field(
        description="The complete set of mutations that can be used in generation."
    )
    alias: str = Field(
        default="combinatorial_generator",
        description="Custom alias to refer to this generator within the task. Useful for tracking provenance of assemblies generated by this generator. For example, this name will appear in the `generator_provenance` column for rows in `selected_sequences` which were produced by this generator.",
    )


class Constraint(BaseModel):
    measure: Measure = Field(description="The measure to constrain.")
    threshold: float | RelativeTo = Field(
        description="The threshold for the constraint, either an absolute value or relative to a reference."
    )


class DiversifyParameters(ParametersBase):
    task_type: ClassVar[str] = "diversify/v2"

    generator: Sequential | Parallel | ModelBasedGenerator | Explicit | CombinatorialGenerator = Field(
        description="The generator configuration or configurations that define requirements and strategies for generating candidate assemblies.",
        discriminator="type",
    )
    ranker: LikelihoodBasedRanker = Field(
        description="The configuration defining how candidate assemblies are evaluated for inclusion in `selected_assemblies`."
    )
    num_assemblies: int = Field(
        description="The target number of `selected_assemblies` to output. In rare cases, the actual number of output assemblies may be slightly lower."
    )


class DiversifyResult(ResultBase):
    selected_assemblies: TableResult = Field(
        description="Table containing the selected assemblies output by `diversify`."
    )
    report: ArtifactParam | None = Field(description="A report on the result of the task.")


class EngineerParameters(ParametersBase):
    task_type: ClassVar[str] = "engineer/v2"

    generator: Sequential | Parallel | ModelBasedGenerator | Explicit | CombinatorialGenerator = Field(
        description="The generator configuration or configurations that define requirements and strategies for generating candidate assemblies.",
        discriminator="type",
    )
    ranker: ScoreBasedRanker = Field(
        description="The configuration defining how candidate assemblies are evaluated for inclusion in `selected_assemblies`."
    )
    num_assemblies: int = Field(
        description="The target number of selected assemblies to output. In rare cases of very difficult `constraints` this task may produce a smaller number."
    )


class EngineerResult(ResultBase):
    selected_assemblies: TableResult = Field(
        description="Table containing the selected assemblies output by `engineer`."
    )
    report: ArtifactParam | None = Field(description="A report on the result of the task.")


class Explicit(BaseModel):
    type: Literal["EXPLICIT"] = Field(default="EXPLICIT")
    source: MonomerSource | VhVlSource
    alias: str = Field(
        default="explicit_generator",
        description="Custom alias to refer to this generator within the task. Useful for tracking provenance of assemblies generated by this generator. For example, this name will appear in the `generator_provenance` column for rows in `selected_sequences` which were produced by this generator.",
    )


class LikelihoodBasedRanker(BaseModel):
    type: Literal["LIKELIHOOD_BASED"] = Field(default="LIKELIHOOD_BASED")
    sampler_config: MonomerSamplerConfig | VhVlSamplerConfig = Field(
        description="The trained sampler artifact used to compute assembly likelihoods, along with any optional control assemblies to include in the final output for experimental validation.",
        discriminator="kind",
    )
    hit_redundancy: int | None = Field(
        default=None,
        description='The maximum number of similar variants within the proposed set.\n\nWe say two variants are "similar" when their predicted performances are highly correlated.\nAt lower values this will mean on average fewer hits, but greater diversity within the set,\nwhile at maximum value (`hit_redundancy=num_selected`) the selection is purely greedy, with\nno requirement on diversity within the set. If set to None, it will be automatically set\nto a default value of `1 + num_assemblies//20`.',
    )


class Measure(BaseModel):
    assay_id: str = Field(description="The ID of the assay.")
    direction: OptimizationDirection = Field(description="The direction of optimization for the assay.")


class ModelBasedGenerator(BaseModel):
    type: Literal["MODEL_BASED"] = Field(default="MODEL_BASED")
    template: MonomerSamplingTemplate | VhVlSamplingTemplate = Field(
        description="The template sequence to use for generation.", discriminator="kind"
    )
    requirements: list[Requirement] = Field(
        default_factory=list,
        description="A list of requirements to satisfy. All requirements must be met for a valid assembly. See the `Requirement` documentation for details on how requirements are evaluated.",
    )
    discourage_mutations: RangeAnnotation | None = Field(
        default=None, description="Expression defining regions where mutations should be discouraged."
    )
    min_mutations: int = Field(description="Minimum number of mutations from the template assembly.")
    max_mutations: int = Field(description="Maximum number of mutations from the template assembly.")
    alias: str = Field(
        default="model_based_generator",
        description="Custom alias to refer to this generator within the task. Useful for tracking provenance of assemblies generated by this generator. For example, this name will appear in the `generator_provenance` column for rows in `selected_sequences` which were produced by this generator.",
    )


class MonomerAssembly(BaseModel):
    monomer: str = Field(description="The monomer sequence")


class MonomerData(BaseModel):
    kind: Literal["MONOMER"] = Field(default="MONOMER")
    homologs: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(description="The homologous assemblies used for training.")
    dataset: (
        list
        | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
        | None
    ) = Field(
        description="The assayed assemblies used for training. If None, predictor finetuning and generator conditioning will be skipped."
    )


class MonomerModels(BaseModel):
    kind: Literal["MONOMER"] = Field(default="MONOMER")
    sampler: ArtifactParam = Field(description="The monomer sampler model.")
    scorer: ArtifactParam | None = Field(description="The monomer scorer model.")
    conditioned_sampler: ArtifactParam | None = Field(
        description="The monomer sampler model conditioned on the primary objective."
    )


class MonomerSamplerConfig(BaseModel):
    kind: Literal["MONOMER_SAMPLER"] = Field(default="MONOMER_SAMPLER")
    sampler: ArtifactParam = Field(description="The trained sampler artifact used to compute sequence likelihoods.")
    reference_assembly: MonomerAssembly = Field(
        description="The reference assembly used as the denominator when computing sequence likelihood ratios."
    )
    controls: (
        list
        | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
        | None
    ) = Field(
        default=None,
        description="Optional control assemblies to include in the final output for experimental validation. These controls should generally be included in the assayed plate, as `selected_assemblies` will be intentionally diverse from these assemblies. However these assemblies will *not* be included in the `selected_assemblies` output table, so -- if used -- take care to re-include them. ",
    )


class MonomerSamplingTemplate(BaseModel):
    kind: Literal["MONOMER"] = Field(default="MONOMER")
    assembly: MonomerAssembly = Field(description="The template sequence to use for generation.")
    sampler: ArtifactParam = Field(description="The trained sampler artifact used to generate candidate sequences.")


class MonomerScorerConfig(BaseModel):
    kind: Literal["MONOMER_SCORER"] = Field(default="MONOMER_SCORER")
    scorer: ArtifactParam = Field(description="The trained scoring model artifact used to score sequences.")
    controls: (
        list
        | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
        | None
    ) = Field(
        default=None,
        description="Optional control sequences to include in the final output for experimental validation. These controls should generally be included in the assayed plate, as `selected_sequences` will be intentionally diverse from these sequences. However these sequences will *not* be included in the `selected_sequences` output table, so -- if used -- take care to re-include them. ",
    )


class MonomerSource(BaseModel):
    kind: Literal["MONOMER"] = Field(default="MONOMER")
    source: list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]


class OASDatabaseSearch(BaseModel):
    type_: Literal["OAS_SEARCH"] = Field(default="OAS_SEARCH", alias="type")
    seed_assemblies: list[VhVlAssembly] = Field(description="The list of light/heavy chain assemblies to search for.")
    search_unpaired: bool = Field(default=True, description="Whether to search for unpaired heavy/light chains.")
    search_paired: bool = Field(default=True, description="Whether to search for paired heavy/light chains.")
    annotate_assembly_features: bool = Field(
        default=True, description="Whether to infer CDR features based on OAS annotations."
    )


class Parallel(BaseModel):
    type: Literal["PARALLEL"] = Field(default="PARALLEL")
    ensemble: list[
        Annotated[
            Sequential | Parallel | ModelBasedGenerator | Explicit | CombinatorialGenerator,
            Field(description="The generation strategy for creating candidate sequences.", discriminator="type"),
        ]
    ] = Field(default_factory=list, description="The list of generators to run in parallel within this ensemble.")
    alias: str | None = Field(
        default=None,
        description="Optional alias to refer to this generator within the task. Useful for tracking provenance of assemblies generated by this generator. If provided, this name will appear in the `generator_provenance` column for rows in `selected_sequences` which were produced by this generator.",
    )


class PreRanked(BaseModel):
    type: Literal["PRE_RANKED"] = Field(default="PRE_RANKED")
    hit_redundancy: int = Field(
        default=3,
        description='The maximum number of similar variants within the proposed set.\n\nWe say two variants are "similar" when their predicted performances are highly correlated.\nAt lower values this will mean on average fewer hits, but greater diversity within the set,\nwhile at maximum value (`hit_redundancy=num_assemblies`) the selection is purely greedy, with\nno requirement on diversity within the set.',
    )


class PrimaryObjective(BaseModel):
    measure: Measure = Field(description="The measure to optimize.")
    reference: float | RelativeTo = Field(
        description="The reference relative to which the primary objective is defined. For multiplicative assays, an absolute reference has to be positive."
    )


class RangeAnnotation(BaseModel):
    ranges: dict[str, list[tuple[int, int]]] = Field(
        description="A mapping from polymer names to annotation ranges. Each range is a tuple of (start, end). Indices are 0-based and inclusive for the start and exclusive for the end. For example, (0, 5) refers to positions 0 to 4."
    )


class RelativeTo(BaseModel):
    reference: MonomerAssembly | VhVlAssembly = Field(description="The reference assembly.")
    margin: float | None = Field(
        default=None,
        description="The desired improvement over the `reference` assembly.\n\nRegardless of the `OptimizationDirection` of the `Measure` that this `RelativeTo` is referring to, larger numbers always mean higher desired improvements.\n\nFor additive assays, a positive margin corresponds to an improvement over the `reference`, and a negative margin to a reduction in performance. For multiplicative assays, the margin is the expected fold improvement, with values greater than 1.0 corresponding to an improvement over the `reference`, and values between 0.0 and 1.0 corresponding to a reduction in performance; negative values are not allowed.\n\nMathematically, the constraint is computed as follows for different scale types and optimization directions:\n\n| Scale Type       | MAXIMIZE | MINIMIZE |\n| ---------------- | -------------------------------- | -------------------------------- |\n| Additive         | `reference + margin`             | `reference - margin`             |\n| Multiplicative   | `reference * margin`             | `reference / margin`             |\n\nFor example, a 10% improvement for a multiplicative assay corresponds to a margin of 1.1.",
    )


class Requirement(BaseModel):
    apply_to: RangeAnnotation
    motifs: list[Annotated[str, Field(description="A motif or regular expression of amino acids.")]]
    operator: Literal["none", "all", "any"] = Field(
        description="The choice of how to interpret the `motifs` together as a requirement."
    )


class ScoreBasedRanker(BaseModel):
    type: Literal["SCORE_BASED"] = Field(default="SCORE_BASED")
    primary_objective: PrimaryObjective = Field(description="The primary optimization objective of `engineer`.")
    constraints: list[Constraint] = Field(
        default_factory=list, description="List of constraints `engineer` will attempt to satisfy."
    )
    scorer_config: MonomerScorerConfig | VhVlScorerConfig = Field(
        description="The trained scoring model artifact used to score sequences, along with any optional control sequences to include in the final output for experimental validation.",
        discriminator="kind",
    )
    hit_redundancy: int | None = Field(
        default=None,
        description='The maximum number of similar variants within the proposed set.\n\nWe say two variants are "similar" when their predicted performances are highly correlated.\nAt lower values this will mean on average fewer hits, but greater diversity within the set,\nwhile at maximum value (`hit_redundancy=num_selected`) the selection is purely greedy, with\nno requirement on diversity within the set. If set to None, it will be automatically set\nto a default value of `1 + num_assemblies//20`.',
    )


class SearchParameters(ParametersBase):
    task_type: ClassVar[str] = "search/v2"

    source: UnirefDatabaseSearch | OASDatabaseSearch = Field(
        description="The source database to search against, along with the assemblies to search for, and database-specific configuration.",
        discriminator="type_",
    )


class SearchResult(ResultBase):
    assembly_features: list[dict[str, AssemblyFeature]] = Field(
        default_factory=list,
        description="Parallel array (wrt to `seed_assemblies`) of features, keyed by feature name.",
    )
    homologs: TableResult
    report: ArtifactParam | None = Field(default=None, description="A report on the result of the task.")


class SelectParameters(ParametersBase):
    task_type: ClassVar[str] = "select/v2"

    assemblies: MonomerSource | VhVlSource = Field(description="The source of assemblies to select from.")
    ranker: (
        Annotated[
            LikelihoodBasedRanker | ScoreBasedRanker,
            Field(description="The ranking strategy for selecting candidate assemblies.", discriminator="type"),
        ]
        | PreRanked
    ) = Field(
        description="The configuration defining how candidate assemblies are evaluated for inclusion in `selected_assemblies`."
    )
    min_assemblies_to_select: int = Field(
        description="The minimum number of assemblies to select. The number of selected candidates is dynamically determined and may be greater than this number. If an exact number of assemblies is desired, set `min_assemblies_to_select` and `max_assemblies_to_select` to the same value."
    )
    max_assemblies_to_select: int | None = Field(
        default=None,
        description="An optional (inclusive) upper limit on the number of assemblies to select. If provided, the task will select between `min_assemblies_to_select` and `max_assemblies_to_select` assemblies.",
    )


class SelectResult(ResultBase):
    selected_assemblies: TableResult = Field(
        description="Table containing the selected assemblies output by `select` sorted in descending order of preference."
    )


class Sequential(BaseModel):
    type: Literal["SEQUENTIAL"] = Field(default="SEQUENTIAL")
    ensemble: list[
        Annotated[
            Sequential | Parallel | ModelBasedGenerator | Explicit | CombinatorialGenerator,
            Field(description="The generation strategy for creating candidate sequences.", discriminator="type"),
        ]
    ] = Field(default_factory=list, description="The list of generators to run in sequence within this ensemble.")
    alias: str | None = Field(
        default=None,
        description="Optional alias to refer to this generator within the task. Useful for tracking provenance of assemblies generated by this generator. If provided, this name will appear in the `generator_provenance` column for rows in `selected_sequences` which were produced by this generator.",
    )


class TrainParameters(ParametersBase):
    task_type: ClassVar[str] = "train/v2"

    data: MonomerData | VhVlData = Field(
        description="The training data, including homologs and assayed sequences.", discriminator="kind"
    )
    assays: list[Assay] = Field(
        default_factory=list,
        description="The assays used for sampler conditioning (on the `primary_objective`) and scorer finetuning and evaluation. Must correspond to assays present in `data.dataset`. If `data.dataset` is None, `assays` must be empty. Otherwise, it must contain at least one assay.",
    )
    primary_objective: Measure | None = Field(
        default=None,
        description="The assay on which the sampler model is conditioned. Should correspond to an assay in `dataset`. If `data.dataset` is None, `primary_objective` must not be set. Otherwise, it must be set.",
    )


class TrainResult(ResultBase):
    models: MonomerModels | VhVlModels = Field(description="The trained models.", discriminator="kind")
    report: ArtifactParam | None = Field(description="A report on the result of the task.")


class UnirefDatabaseSearch(BaseModel):
    type_: Literal["UNIREF_SEARCH"] = Field(default="UNIREF_SEARCH", alias="type")
    seed_assemblies: list[MonomerAssembly] = Field(description="A list of monomeric assemblies to search for.")


class VhVlAssembly(BaseModel):
    vh: str | None = Field(default=None, description="The heavy chain sequence.")
    vl: str | None = Field(default=None, description="The light chain sequence.")


class VhVlData(BaseModel):
    kind: Literal["VHVL"] = Field(default="VHVL")
    homologs: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(description="The homologous assemblies used for training.")
    dataset: (
        list
        | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
        | None
    ) = Field(
        description="The assayed assemblies used for training. If None, predictor finetuning and generator conditioning will be skipped."
    )


class VhVlModels(BaseModel):
    kind: Literal["VH_VL"] = Field(default="VH_VL")
    sampler: ArtifactParam = Field(description="The VhVl sampler model.")
    scorer: ArtifactParam | None = Field(description="The VhVl scorer model.")
    conditioned_sampler: ArtifactParam | None = Field(
        description="The VhVl sampler model conditioned on the primary objective."
    )


class VhVlSamplerConfig(BaseModel):
    kind: Literal["VHVL_SAMPLER"] = Field(default="VHVL_SAMPLER")
    sampler: ArtifactParam = Field(description="The trained sampler artifact used to compute assembly likelihoods.")
    reference_assembly: VhVlAssembly = Field(
        description="The reference assembly used as the denominator when computing assembly likelihood ratios."
    )
    controls: (
        list
        | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
        | None
    ) = Field(
        default=None,
        description="Optional control assemblies to include in the final output for experimental validation. These controls should generally be included in the assayed plate, as `selected_assemblies` will be intentionally diverse from these assemblies. However these assemblies will *not* be included in the `selected_assemblies` output table, so -- if used -- take care to re-include them. ",
    )


class VhVlSamplingTemplate(BaseModel):
    kind: Literal["VHVL"] = Field(default="VHVL")
    assembly: VhVlAssembly = Field(description="The template sequence to use for generation.")
    sampler: ArtifactParam = Field(description="The trained sampler artifact used to generate candidate sequences.")


class VhVlScorerConfig(BaseModel):
    kind: Literal["VHVL_SCORER"] = Field(default="VHVL_SCORER")
    scorer: ArtifactParam = Field(description="The trained scoring model artifact used to score sequences.")
    controls: (
        list
        | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
        | None
    ) = Field(
        default=None,
        description="Optional control sequences to include in the final output for experimental validation. These controls should generally be included in the assayed plate, as `selected_sequences` will be intentionally diverse from these sequences. However these sequences will *not* be included in the `selected_sequences` output table, so -- if used -- take care to re-include them. ",
    )


class VhVlSource(BaseModel):
    kind: Literal["VHVL"] = Field(default="VHVL")
    source: list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]



from __future__ import annotations

from abc import ABC
from typing import TypeAlias

from pydantic import BaseModel, Field

from .common import (
    ArchivableResourceMixin,
    BaseListResponse,
    ContextProject,
    ContextRound,
    ContextWorkspace,
    ErrorResponseMixin,
    ResourceResponse,
)
from .tasks import v1, v2
from .tasks.common import TaskState


class ListTaskResponse(BaseListResponse):
    items: list[TaskResponse]


class TaskUpdateBase(BaseModel, ABC):
    display_name: str | None = Field(
        default=None, description="Optional display name for the task. This name is mutable."
    )
    description: str | None = Field(
        default=None, description="Optional description to provide additional notes about the executed task."
    )


class TaskResponse(ResourceResponse, ArchivableResourceMixin, ErrorResponseMixin):
    type: str
    context: ContextWorkspace | ContextProject | ContextRound = Field(
        description="The context in which this task runs. Tasks started in a project or round context will only 'see' table data that belongs to the associated project.",
        discriminator="kind",
    )
    name: str | None = Field(description="An optional name of the task.", deprecated="Use display_name instead.")
    display_name: str | None = Field(default=None, description="An optional display name of the task.")
    idempotency_key: str | None = Field(
        default=None,
        description="An optional idempotency key of the task. It must be unique within the task's context.",
    )
    description: str | None = Field(description="A description of the task.")
    data_version_id: int | None = Field(description="The data version at which table inputs are observed.")
    data_load_id: int | None = Field(description="ID of the data load for the task's results.")
    parameters: (
        v1.AnalyzeDataParameters
        | v1.AnalyzeDiversifyParameters
        | v1.AnalyzeEngineerParameters
        | v1.AnalyzeTrainParameters
        | v1.DiversifyParameters
        | v1.EngineerParameters
        | v1.SearchParameters
        | v1.SelectParameters
        | v1.TrainParameters
        | v2.AnalyzeDataParameters
        | v2.DiversifyParameters
        | v2.EngineerParameters
        | v2.SelectParameters
        | v2.SearchParameters
        | v2.TrainParameters
    ) = Field(description="The parameters of the task.")
    result: (
        v1.AnalyzeResult
        | v1.DiversifyResult
        | v1.EngineerResult
        | v1.SearchResult
        | v1.SelectResults
        | v1.TrainResult
        | v2.AnalyzeDataResult
        | v2.DiversifyResult
        | v2.EngineerResult
        | v2.SelectResult
        | v2.SearchResult
        | v2.TrainResult
        | None
    ) = Field(description="The result of the completed task.")
    state: TaskState = Field(description="The current state of the task's execution.")


class TaskCreate(TaskUpdateBase):
    parameters: (
        v1.AnalyzeDataParameters
        | v1.AnalyzeDiversifyParameters
        | v1.AnalyzeEngineerParameters
        | v1.AnalyzeTrainParameters
        | v1.DiversifyParameters
        | v1.EngineerParameters
        | v1.SearchParameters
        | v1.SelectParameters
        | v1.TrainParameters
        | v2.AnalyzeDataParameters
        | v2.DiversifyParameters
        | v2.EngineerParameters
        | v2.SelectParameters
        | v2.SearchParameters
        | v2.TrainParameters
    ) = Field(description="The parameters for the task.")
    context: ContextWorkspace | ContextProject | ContextRound = Field(
        description="The context in which the task is executed.", discriminator="kind"
    )
    data_version_id: int | None = Field(
        default=None,
        description="For tasks that have parameters referencing data tables, this specifies the version of the table data to use. If no version is specified, it defaults to the most recent data version.",
    )
    name: str | None = Field(
        default=None,
        description="Optional name for the task. If provided it will ensure that only one task with that name exists within the task's context. This field is deprecated in favor of display_name + idempotency_key.",
    )
    idempotency_key: str | None = Field(
        default=None,
        description="Optional idempotency key to ensure only one task with this key exists within its context. If provided and a task with this key already exists, the existing task will be returned instead of creating a new one.",
    )


class TaskUpdate(TaskUpdateBase):
    pass


TaskParameters: TypeAlias = (
    v1.AnalyzeDataParameters
    | v1.AnalyzeDiversifyParameters
    | v1.AnalyzeEngineerParameters
    | v1.AnalyzeTrainParameters
    | v1.DiversifyParameters
    | v1.EngineerParameters
    | v1.SearchParameters
    | v1.SelectParameters
    | v1.TrainParameters
    | v2.AnalyzeDataParameters
    | v2.DiversifyParameters
    | v2.EngineerParameters
    | v2.SelectParameters
    | v2.SearchParameters
    | v2.TrainParameters
)
TaskResults: TypeAlias = (
    v1.AnalyzeResult
    | v1.DiversifyResult
    | v1.EngineerResult
    | v1.SearchResult
    | v1.SelectResults
    | v1.TrainResult
    | v2.AnalyzeDataResult
    | v2.DiversifyResult
    | v2.EngineerResult
    | v2.SelectResult
    | v2.SearchResult
    | v2.TrainResult
)



from __future__ import annotations

from enum import StrEnum

from pydantic import BaseModel, Field

from .common import ArchivableResourceMixin, BaseListResponse, ErrorResponseMixin, ResourceResponse


class WorkspaceState(StrEnum):
    PROVISIONING = "PROVISIONING"
    DELETING = "DELETING"
    READY = "READY"
    DELETED = "DELETED"
    FAILED = "FAILED"


class ListProjectResponse(BaseListResponse):
    items: list[ProjectResponse]


class ListRoundResponse(BaseListResponse):
    items: list[RoundResponse]


class ProjectCreate(BaseModel):
    name: str


class ProjectResponse(ResourceResponse, ArchivableResourceMixin):
    name: str


class RoundCreate(BaseModel):
    project_id: int = Field(description="The project to which the round belongs")
    name: str = Field(description="The name of the round")
    description: str | None = Field(default=None)


class RoundResponse(ResourceResponse, ArchivableResourceMixin):
    project_id: int
    name: str
    description: str | None


class WorkspaceResponse(ResourceResponse, ErrorResponseMixin):
    name: str
    display_name: str
    state: WorkspaceState
    picture_file_id: int | None = Field(default=None)



"""Cradle Platform API SDK

This SDK provides a convenient Python interface for interacting with the Cradle Platform APIs.
It allows you to programmatically access and manage your data, projects, rounds, and tasks within the Cradle platform.

Key Features:
- Seamless OAuth authentication with the Cradle Platform via device flow.
- Easy access to various client interfaces for different API functionalities (Data, Project, Round, Task).
- Simplified data upload and management.

Getting Started:
1.  Installation:
    Install the SDK using pip::

        pip install cradle-sdk

2.  Authentication:
    The SDK uses OAuth device flow for authentication. Simply instantiate the client,
    and it will provide you with a URL to complete authentication in your browser.

3.  Basic Usage:
    Initialize the client and follow the authentication prompt:
    ```python
    from cradle.sdk import Client

    # Initialize the client - this will prompt for authentication
    client = Client(workspace="YOUR_WORKSPACE_NAME")

    # The client will display a message like this:
    # Please click the following URL to complete authentication
    # https://signin.cradle.bio/device?user_code=XXXX-XXXX
    # and verify that the code matches: XXXX-XXXX

    # After authenticating in your browser via SSO or Cradle credentials,
    # the client will be ready to use.

    # Access specific API clients
    data_client = client.data
    project_client = client.project
    round_client = client.round
    task_client = client.task

    # Example: List projects
    projects = project_client.list_projects()
    for project in projects:
        print(project.name)
    ```

For detailed documentation on each module and class, please refer to the respective sections.
"""

from .client import (
    Client as Client,
    DataClient as DataClient,
    DataLoadClient as DataLoadClient,
    ProjectClient as ProjectClient,
    RoundClient as RoundClient,
    TaskClient as TaskClient,
)
from .exceptions import ClientError as ClientError



from __future__ import annotations

from enum import StrEnum
from typing import Annotated, ClassVar, Literal

from pydantic import BaseModel, Field

from .common import (
    ArtifactParam,
    ParametersBase,
    ResultBase,
    TableResult,
    TableSourceQuery,
    TableSourceTable,
    TableSourceTaskResult,
)


class FeatureAnnotation(StrEnum):
    """Annotation to be configured per residue in a domain:

    UNRELIABLE: The evolutionary context is not reliable, eg in CDRs of antibodies.
    IGNORE: A part of the sequence that should not be used for evolutionary search and won't be mutated, this can be tags, linkers or other artificial subsequences.
    """

    UNRELIABLE = "UNRELIABLE"
    IGNORE = "IGNORE"


class OptimizationDirection(StrEnum):
    MAXIMIZE = "MAXIMIZE"
    MINIMIZE = "MINIMIZE"


class ProteinDatabase(StrEnum):
    UNIREF_30 = "UNIREF_30"
    OAS_90 = "OAS_90"


class ProteinModelType(StrEnum):
    DEFAULT = "DEFAULT"
    ANTIBODY = "ANTIBODY"


class ScaleType(StrEnum):
    MULTIPLICATIVE = "MULTIPLICATIVE"
    ADDITIVE = "ADDITIVE"
    RANK = "RANK"


class AbsoluteConstraint(BaseModel):
    type_: Literal["ABSOLUTE_CONSTRAINT"] = Field(default="ABSOLUTE_CONSTRAINT", alias="type")
    assay_id: str = Field(description="The ID of the constrained assay.")
    direction: OptimizationDirection = Field(
        description="Whether the assay value should be above (MAXIMIZE) or below (MINIMIZE) the threshold."
    )
    threshold: float = Field(description="The threshold assay value.")


class AnalyzeDataParameters(ParametersBase):
    task_type: ClassVar[str] = "analyze.data/v1"

    reference_sequence: str = Field(description="The sequence to consider as a reference for computing mutations.")
    dataset: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(description="The assay data to be used for model training.")
    assays: list[Assay] = Field(default_factory=list, description="List of assay metadata entries")


class AnalyzeDiversifyParameters(ParametersBase):
    task_type: ClassVar[str] = "analyze.diversify/v1"

    selected_sequences: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(description="The name of the input table with the selected sequences.")
    generated_sequences: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(description="The name of the input view/table with the generated sequences.")
    template_metadata: TemplateMetadata = Field(description="Metadata about templates.")


class AnalyzeEngineerParameters(ParametersBase):
    task_type: ClassVar[str] = "analyze.engineer/v1"

    selected_sequences: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(description="The name of the input table with the selected sequences.")
    engineered_sequences: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(description="The name of the input view/table with the generated sequences.")
    assays: list[Assay] = Field(
        default_factory=list, description="The list of objectives that were optimized during training."
    )
    primary_objective: PrimaryObjective = Field(description="The primary objective to be optimized.")
    constraints: list[Annotated[AbsoluteConstraint | RelativeConstraint, Field(discriminator="type_")]] = Field(
        default_factory=list, description="The constraints applied to engineering."
    )
    template_metadata: list[TemplateMetadata] = Field(description="Metadata about templates.")
    new_mutation_ratio: float = Field(description="Ratio of new mutations in the generated sequences.")
    scorer: ArtifactParam = Field(description="Scoring model used to predict assay values for candidate sequences.")


class AnalyzeResult(ResultBase):
    report: ArtifactParam = Field(description="The report data")


class AnalyzeTrainParameters(ParametersBase):
    task_type: ClassVar[str] = "analyze.train/v1"

    prediction_data: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(description="Table of predicted assay values for the input dataset under the scorer.")
    generator_prediction_data: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(description="Pseudo log-likelihoods of the sequences in the dataset under the sampler.")
    assays: list[Assay] = Field(default_factory=list, description="List of assay metadata entries")
    primary_objective: PrimaryObjective = Field(
        description="The primary objective on which the sampler model is conditioned."
    )
    constraints: list[Annotated[AbsoluteConstraint | RelativeConstraint, Field(discriminator="type_")]] = Field(
        default_factory=list, description="The constraints applied to training."
    )


class Assay(BaseModel):
    assay_id: str = Field(description="ID of the assay")
    name: str = Field(description="Human readable name of the assay")
    scale_type: ScaleType = Field(
        description="The scale type defines how the assay behaves.\n\n* **Additive**: Assay values are comparable in magnitude over batches, a delta of eg 5 in one batch is equivalent to 5 in another batch.\n* **Multiplicative**: Assay values are comparable over batches by applying a multiplication factor, eg fold improvement: `score_variantB_batch1 = factor(typically starting sequence score) * score_variantB_batch`.\n* **Rank**: Assay values are not comparable over batches, we can only assume the ranking within a batch is correct."
    )
    unit: str | None = Field(default=None, description="Unit of the assay - used for display purposes only")


class BlockedAAItem(BaseModel):
    blocked_aas: str = Field(
        description="One or more amino acid letter code to be blocked from being mutated to. Use '*' to indicate that a position is fully blocked from being mutated."
    )
    ranges: list[tuple[int, int]] = Field(
        description="List of ranges where the mutations are blocked. Each range is a tuple of (start, end). Indices are 0-based and inclusive for the start and exclusive for the end. For example, (0, 5) blocks positions 0 to 4."
    )


class BlockedMotifItem(BaseModel):
    blocked_motif: str = Field(description="Amino acid motif to be blocked from appearing in the final sequence.")
    ranges: list[tuple[int, int]] = Field(
        description="List of ranges where the motif is blocked. Each range is a tuple of (start, end). Indices are 0-based and inclusive for the start and exclusive for the end. For example, (0, 5) blocks positions 0 to 4. The entire motif must fit into the specified positions to be blocked."
    )


class ClusterParameters(BaseModel):
    e_value_threshold: float = Field(
        default=float("inf"),
        description="Maximum e-value allowed between linked sequences in clustering algorithm. A lower number will result in tighter clusters.",
    )
    coverage_threshold: float = Field(
        default=0.8,
        description="Minimum coverage allowed between linked sequences in clustering algorithm. A higher number will result in tighter clusters.",
    )
    sequence_identity_threshold: float = Field(
        default=0.0,
        description="Minimum sequence identity allowed between linked sequences in clustering algorithm. A higher number will result in tighter clusters.",
    )


class ClusteredBatches(BaseModel):
    assayed_sequences: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(description="Table containing assayed sequences. Everything outside of this set will not be selected.")
    sort_parameters: SortParameters
    filter_parameters: FilterParameters | None = Field(default=None, description="If `None`, no filtering is applied.")
    cluster_ids_or_parameters: (
        list
        | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
        | ClusterParameters
    ) = Field(
        default_factory=ClusterParameters,
        description="Either a table mapping sequences to cluster ids or parameters for running clustering.",
    )


class DatabaseSearch(BaseModel):
    type_: Literal["DATABASE_SEARCH"] = Field(default="DATABASE_SEARCH", alias="type")
    seed_domains: list[str] = Field(description="A list of domain sequences to search for.")
    database: ProteinDatabase = Field(
        default=ProteinDatabase.UNIREF_30, description="Type of the protein database to use."
    )


class DiversifyParameters(ParametersBase):
    task_type: ClassVar[str] = "diversify/v1"

    template_sequence: DiversifyTemplateInputs = Field(description="The template amino acid sequence to diversify.")
    homologs: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(
        description="Table of homologs to use for generator training. In the simplest case, this is just the result of a multiple sequence alignment (MSA) against the sequence to optimize."
    )
    domain_features: dict[
        str,
        Annotated[
            list[DomainFeatureItem],
            Field(
                description="List of domain annotations. A residue may be assigned zero or one annotations. Ranges which are not annotated are considered reliable, i.e. the MSA in those ranges is used to infer conservation-based features."
            ),
        ],
    ] = Field(
        default_factory=dict,
        description="Properties for each domain. A domain is one or more subsequences of the protein sequence, for example the active site of an enzyme, CDRs of an antibody, heavy and light chains of an scFv, etc. ",
    )
    protein_model_type: ProteinModelType = Field(
        default=ProteinModelType.DEFAULT,
        description="Specifies the type of base model to use for training: use `DEFAULT` for single-chain proteins and `ANTIBODY` for multi-chain antibodies.",
    )


class DiversifyResult(ResultBase):
    selected_sequences: TableResult = Field(description="Table containing the selected sequences.")
    generated_sequences: TableResult = Field(description="Table containing the generated sequences.")
    template_metadata: TemplateMetadata = Field(description="Metadata about templates.")
    report: ArtifactParam = Field(description="A report on the result of the task.")


class DiversifyTemplateInputs(BaseModel):
    template_id: str = Field(
        description="Identifier for the template used to identify the source of generated sequences in the output."
    )
    sequence: str = Field(description="The template amino acid sequence")
    num_results: int = Field(
        description="Number of sequences to generate for the final plate from the template sequence."
    )
    min_mutations: int = Field(default=1, description="Minimum number of mutations per sequence")
    max_mutations: int = Field(default=4, description="Maximum number of mutations per sequence")
    blocked_aas: list[BlockedAAItem] = Field(
        default_factory=list,
        description="List of blocked amino acids which may not be mutated to, with their respective ranges.",
    )
    blocked_motifs: list[BlockedMotifItem] = Field(
        default_factory=list, description="List of blocked amino acid motifs with their respective ranges."
    )


class DomainFeatureItem(BaseModel):
    annotation: FeatureAnnotation = Field(
        description="Annotation for the residues. `UNRELIABLE`: do not look at the MSA to infer conservation-based features (e.g. CDRs) for the protein `IGNORE`: do not change this position at all (e.g. for linkers) and do not use to infer any conservation-based features."
    )
    ranges: list[tuple[int, int]] = Field(
        description="List of ranges where the annotation applies. Each range is a tuple of (start, end). Indices are 0-based and inclusive for the start and exclusive for the end. For example, (0, 5) blocks positions 0 to 4."
    )


class EngineerParameters(ParametersBase):
    task_type: ClassVar[str] = "engineer/v1"

    dataset: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(description="The table of assayed data, to be used for exploitation.")
    assays: list[Assay] = Field(default_factory=list, description="List of assay metadata entries")
    primary_objective: PrimaryObjective = Field(description="The primary objective the samplers are trained for.")
    constraints: list[Annotated[AbsoluteConstraint | RelativeConstraint, Field(discriminator="type_")]] = Field(
        default_factory=list, description="List of assay constraints"
    )
    samplers: list[ArtifactParam] = Field(description="The sampling models used to generate candidate sequences.")
    scorer: ArtifactParam = Field(description="Scoring model used to predict assay values for candidate sequences.")
    template_sequences: list[TemplateInputs] = Field(description="Initial sequences used as templates for generation")


class EngineerResult(ResultBase):
    selected_sequences: TableResult = Field(description="Table containing the selected sequences.")
    engineered_sequences: TableResult = Field(description="Table containing the generated sequences.")
    num_evaluated_seqs: int = Field(
        description="Number of sequences evaluated in silico (only the ones sent to the predictor)."
    )
    num_generated_seqs: int = Field(
        description="Number of sequences generated in silico (this is > num_evaluated_seqs, as it includes generated sequences that were not sent to the predictor, due to lower likelihood)."
    )
    new_mutation_ratio: float = Field(
        description="Number of mutations in `selected_candidates` that do not appear in train/test datasets."
    )
    template_metadata: list[TemplateMetadata] = Field(description="Metadata about templates.")
    report: ArtifactParam = Field(description="A report on the result of the task.")


class FilterParameters(BaseModel):
    objective: PrimaryObjective = Field(
        description="The objective used for filtering clusters. Assay id must match one of the assay ids in the input dataset."
    )
    hit_redundancy: int = Field(
        default=5,
        description='The maximum number of similar variants within the proposed set.\n\nWe say two variants are "similar" when their predicted performances are highly correlated.\nAt lower values this will mean on average fewer hits, but greater diversity within the set,\nwhile at maximum value (hit_redundancy=num_selected) the selection is purely greedy, with\nno requirement on diversity within the set.',
    )


class Homologs(BaseModel):
    type_: Literal["HOMOLOGS"] = Field(default="HOMOLOGS", alias="type")
    homologs: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(description="Precomputed homologs table.")


class LabeledSelectionParameters(BaseModel):
    assayed_sequences: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(
        description="Table containing sequences with assays to use for down-selection. Everything outside of this set will not be selected."
    )


class PrimaryObjective(BaseModel):
    assay_id: str = Field(description="The ID of the assay to be optimized.")
    direction: OptimizationDirection = Field(description="Whether the maximize or minimize the assay value.")


class RelativeConstraint(BaseModel):
    type_: Literal["RELATIVE_CONSTRAINT"] = Field(default="RELATIVE_CONSTRAINT", alias="type")
    assay_id: str = Field(description="The ID of the constrained assay.")
    direction: OptimizationDirection = Field(
        description="Whether the assay value of new sequences should be above (MAXIMIZE) or below (MINIMIZE) the assay value of the `relative_to` sequence."
    )
    relative_to: str = Field(description="The sequence relative to which new sequences are constrained.")


class ScoredSelectionParameters(BaseModel):
    scorer: ArtifactParam
    hit_redundancy: int = Field(
        default=10,
        description="Number of top candidates for which to optimize the average score. A higher number means less diversity.",
    )
    max_mutation_frequency: float | None = Field(
        default=None,
        description="Maximum mutation frequency in the set of selected sequences. If None, the algorithm will pick sensible defaults.",
    )


class SearchParameters(ParametersBase):
    task_type: ClassVar[str] = "search/v1"

    source: Homologs | DatabaseSearch = Field(
        description="Precomputed homologs table or seed sequences to search for. If a table of precomputed sequence homologs is provided, the entries in the `seed_domain` column of this table may be used for sequence feature computation. For example, if the sequence to be optimized is an scFv, the `seed_domain` column would contain the heavy and light chain scaffold sequences of the scFv. These 2 scaffold sequences will be used to search for homologs and infer features such as the CDRs.",
        discriminator="type_",
    )
    antibody_sequence_features: bool = Field(
        default=False, description="Whether to compute sequence features. Only supported for antibodies."
    )


class SearchResult(ResultBase):
    domain_features: dict[
        str,
        Annotated[
            list[DomainFeatureItem],
            Field(
                description="List of domain annotations. A residue may be assigned zero or one annotations. Ranges which are not annotated are considered reliable, i.e. the MSA in those ranges is used to infer conservation-based features."
            ),
        ],
    ] = Field(
        default_factory=dict,
        description="Mapping of protein sequences to their respective domain features. Each range in the domain features must be a valid index into the corresponding protein sequence.",
    )
    homologs: TableResult


class SelectParameters(ParametersBase):
    task_type: ClassVar[str] = "select/v1"

    sequences: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(description="The sequences to select from.")
    objective: PrimaryObjective = Field(
        description="The primary objective to optimize during selection. This is used for scoring sequences."
    )
    constraints: list[Annotated[AbsoluteConstraint | RelativeConstraint, Field(discriminator="type_")]] = Field(
        default_factory=list, description="Constraints that must be satisfied by the selected sequences."
    )
    selection_method: ScoredSelectionParameters | LabeledSelectionParameters = Field(
        description="Method of selection. Can either be based on a scorer or on hard labels."
    )
    clustered_batches: ClusteredBatches | None = Field(
        default=None,
        description="Determines the clusters and representatives to use for selection. If `None`, all sequences are passed for selection at once.",
    )
    max_num_sequences_to_select: int = Field(description="No more than this many sequences will be selected.")
    min_num_sequences_to_select: int = Field(
        default=0, description="Warning will be raised if fewer than this many sequences are selected."
    )


class SelectResults(ResultBase):
    selected_sequences: TableResult = Field(description="Table containing the selected sequences.")


class SortParameters(BaseModel):
    objective: PrimaryObjective = Field(
        description="The objective used for sorting clusters. Assay id must match one of the assay ids in the input dataset. Sorting direction is determined by `objective.direction`."
    )


class TemplateInputs(BaseModel):
    template_id: str = Field(
        description="Identifier for the template used to identify the source of generated sequences in the output."
    )
    sequence: str = Field(description="The template amino acid sequence")
    num_results: int = Field(
        description="Number of sequences generated from this template to be added to the final plate."
    )
    min_mutations: int = Field(default=1, description="Minimum number of mutations per sequence.")
    max_mutations: int = Field(default=8, description="Maximum number of mutations per sequence.")
    hit_redundancy: int = Field(
        default=10,
        description="Number of top candidates for which to optimize the average score. A higher number means less diversity.",
    )
    blocked_aas: list[BlockedAAItem] = Field(
        default_factory=list,
        description="List of blocked amino acids which may not be mutated to, with their respective ranges.",
    )
    blocked_motifs: list[BlockedMotifItem] = Field(
        default_factory=list, description="List of blocked amino acid motifs with their respective ranges."
    )
    blocked_regexps: list[str] | None = Field(
        default_factory=list, description="Regular expressions that the sequence cannot match"
    )
    must_match_regexps: list[str] | None = Field(
        default_factory=list,
        description="Generated sequences must match all of the given regular expressions. Can be used to enforce certain motifs.",
    )


class TemplateMetadata(BaseModel):
    template_id: str = Field(
        description="Identifier for the template used to identify the source of generated sequences."
    )
    sequence: str = Field(description="Amino acid template sequence used for generation.")
    blocked_aas: list[BlockedAAItem] = Field(
        default_factory=list,
        description="List of blocked amino acids which may not be mutated to, with their respective ranges.",
    )
    min_mutations: int
    max_mutations: int
    structure: ArtifactParam | None = Field(default=None)


class TrainParameters(ParametersBase):
    task_type: ClassVar[str] = "train/v1"

    homologs: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(description="The table of homologous sequences to be used for training the base model.")
    domain_features: dict[
        str,
        Annotated[
            list[DomainFeatureItem],
            Field(
                description="List of domain annotations. A residue may be assigned zero or one annotations. Ranges which are not annotated are considered reliable, i.e. the MSA in those ranges is used to infer conservation-based features."
            ),
        ],
    ] = Field(
        default_factory=dict,
        description="Mapping of protein sequences to their respective domain features. Each range in the domain features must be a valid index into the corresponding protein sequence.",
    )
    protein_model_type: ProteinModelType = Field(
        default=ProteinModelType.DEFAULT,
        description="Specifies the type of base model to use for training: use `DEFAULT` for single-chain proteins and `ANTIBODY` for multi-chain antibodies.",
    )
    dataset: (
        list | Annotated[TableSourceTable | TableSourceQuery | TableSourceTaskResult, Field(discriminator="kind")]
    ) = Field(description="The name of the input table to be used for model training.")
    assays: list[Assay] = Field(default_factory=list, description="List of assay metadata entries")
    primary_objective: PrimaryObjective = Field(
        description="The primary objective on which the sampler model is conditioned. Should correspond to an assay in `dataset`."
    )


class TrainResult(ResultBase):
    base_sampler: ArtifactParam = Field(
        description="A sampler which attempts to produce plausible sequences as understood relative to the passed `homologs`. "
    )
    conditioned_sampler: ArtifactParam | None = Field(
        description="A sampler which attempts to produce high quality sequences as understood relative to the passed `primary_objective`."
    )
    scorer: ArtifactParam = Field(description="The scoring model, trained on the assay data provided.")
    prediction_data: TableResult = Field(description="Predicted assay values on all folds of the input dataset.")
    generator_prediction_data: TableResult = Field(
        description="Pseudo log-likelihoods of the sequences in the dataset under the generator(s)."
    )
    report: ArtifactParam = Field(description="A report on the result of the task.")



from __future__ import annotations

from enum import StrEnum
from typing import Annotated, Any, Literal, TypeAlias

from pydantic import BaseModel, Field

from .common import (
    ArchivableResourceMixin,
    BaseListResponse,
    ContextProject,
    ContextRound,
    ContextWorkspace,
    ErrorResponseMixin,
    ImmutableResourceResponse,
    ResourceResponse,
)


class ArtifactProducerKind(StrEnum):
    TASK = "TASK"
    USER = "USER"


class DataActionKind(StrEnum):
    LOAD = "LOAD"
    UNDO_LOAD = "UNDO_LOAD"
    TABLE_CREATE = "TABLE_CREATE"
    TABLE_UPDATE = "TABLE_UPDATE"
    TABLE_ARCHIVE = "TABLE_ARCHIVE"


class DataLoadState(StrEnum):
    PENDING = "PENDING"
    LOADING = "LOADING"
    DELETING = "DELETING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"


class FileFormat(StrEnum):
    CSV = "CSV"
    PARQUET = "PARQUET"
    NDJSON = "NDJSON"


class TableKind(StrEnum):
    TABLE = "TABLE"
    VIEW = "VIEW"


class TypeNames(StrEnum):
    BOOL = "BOOL"
    INT64 = "INT64"
    FLOAT64 = "FLOAT64"
    STRING = "STRING"
    STRUCT = "STRUCT"
    ARRAY = "ARRAY"


class AddTableRequest(BaseModel):
    table_ref: str
    config: TableLoadConfig


class ColumnBase(BaseModel):
    name: str
    metadata: ColumnMetadata | None = Field(default=None)


class ArrayType(BaseModel):
    type: Literal[TypeNames.ARRAY] = Field(default=TypeNames.ARRAY)
    item_type: PrimitiveType | StructType | ArrayType = Field(discriminator="type")


class ArtifactResponse(ImmutableResourceResponse):
    context: ContextWorkspace | ContextProject | ContextRound = Field(discriminator="kind")
    type: str
    data: dict[str, Any]
    producer: TaskProducer | UserProducer = Field(discriminator="kind")


class TableCreateBase(BaseModel):
    kind: TableKind
    reference: str = Field(
        description="A unique name to use for referring to the table, such as `temperature_assay`. User-created\ntables should always use this simple `table_name` format. Cradle-created table references have the form\n`catalog.schema.table_name`, for example `cradle.results.engineered_sequences`.\n\nReferences must be unique per workspace. One exception to the rule is that once a table is archived the\nrefence is available to be used again. References are tied to a data version, i.e. when running a query\nat a historic version, a reference will point to the table that held the reference at this data version."
    )
    description: str | None = Field(default=None, description="A human-friendly description of the table's contents.")


class TableResponseBase(ResourceResponse, ArchivableResourceMixin):
    """Base class for table responses."""

    kind: TableKind
    reference: str = Field(
        description="A unique name to use for referring to the table, such as `temperature_assay`. User-created\ntables should always use this simple `table_name` format. Cradle-created table references have the form\n`catalog.schema.table_name`, for example `cradle.results.engineered_sequences`.\n\nReferences must be unique per workspace. One exception to the rule is that once a table is archived the\nrefence is available to be used again. References are tied to a data version, i.e. when running a query\nat a historic version, a reference will point to the table that held the reference at this data version."
    )
    description: str | None
    columns: list[Annotated[PrimitiveColumn | StructColumn | ArrayColumn, Field(discriminator="type")]] = Field(
        description="The columns of the table. For views, the columns and their types are inferred from the query at view creation or update time."
    )
    version_id: int = Field(
        description="The version ID of the table state reflected in the response.Changes to a view's query or a base table's columns cause a version change as well as renames cause a version change."
    )


class _TableUpdateBase(BaseModel):
    description: str | None = Field(default=None, description="Description of the table.")


class ColumnMetadata(BaseModel):
    description: str | None = Field(default=None)
    deprecated: bool = Field(default=False)


class DataActionLoad(BaseModel):
    kind: Literal[DataActionKind.LOAD] = Field(default=DataActionKind.LOAD)
    load_id: int = Field(description="The ID of the data load from which data was added to the dataset.")


class DataActionTableArchive(BaseModel):
    kind: Literal[DataActionKind.TABLE_ARCHIVE] = Field(default=DataActionKind.TABLE_ARCHIVE)
    table_id: int = Field(description="The ID of the table that was archived.")


class DataActionTableCreate(BaseModel):
    kind: Literal[DataActionKind.TABLE_CREATE] = Field(default=DataActionKind.TABLE_CREATE)
    table_id: int = Field(description="The ID of the table that was created.")
    table_version_id: int = Field(description="The ID of the table version that was created.")


class DataActionTableUpdate(BaseModel):
    kind: Literal[DataActionKind.TABLE_UPDATE] = Field(default=DataActionKind.TABLE_UPDATE)
    table_id: int = Field(description="The ID of the table that was updated.")
    table_version_id: int = Field(description="The ID of the new table version.")


class DataActionUndoLoad(BaseModel):
    kind: Literal[DataActionKind.UNDO_LOAD] = Field(default=DataActionKind.UNDO_LOAD)
    load_id: int = Field(description="ID of the data load of which data was removed from the dataset.")


class DataLoadCreate(BaseModel):
    context: ContextProject | ContextRound = Field(
        description="The context the loaded data will be associated with. The data will either live within a Project (visible to all rounds) or within a specific Round of a project (visible only to that round)."
    )
    tables: dict[str, TableLoadConfig] = Field(
        default_factory=dict,
        description="A map of table references to table schemas. The map keys define the tables for which data is uploaded. The map values define the schema and file format for a given table.",
    )


class DataLoadResponse(ResourceResponse, ErrorResponseMixin):
    context: ContextWorkspace | ContextProject | ContextRound = Field(
        description="The context the loaded data is be associated with", discriminator="kind"
    )
    tables: dict[str, TableLoadConfigResponse] = Field(
        default_factory=dict, description="The tables for which data will be loaded."
    )
    state: DataLoadState
    is_active: bool
    files: list[FileUploadResponse] = Field(description="Files that have been uploaded to this data load.")


class DataVersionResponse(ImmutableResourceResponse):
    action: (
        DataActionLoad
        | DataActionUndoLoad
        | Annotated[DataActionTableCreate | DataActionTableUpdate | DataActionTableArchive, Field(discriminator="kind")]
    ) = Field(discriminator="kind")


class FileUploadResponse(ImmutableResourceResponse):
    filepath: str
    size: int
    table_reference: str | None
    source_file_id: int | None
    description: str | None


class ListArtifactResponse(BaseListResponse):
    items: list[ArtifactResponse]


class ListDataLoadResponse(BaseListResponse):
    items: list[DataLoadResponse]


class ListDataVersionResponse(BaseListResponse):
    items: list[DataVersionResponse]


class ListTableResponse(BaseListResponse):
    items: list[Annotated[BaseTableResponse | ViewTableResponse, Field(discriminator="kind")]]


class PrimitiveType(BaseModel):
    type: Literal[TypeNames.BOOL, TypeNames.INT64, TypeNames.FLOAT64, TypeNames.STRING]


class QueryDataRequest(BaseModel):
    query: str = Field(description="The SQL query to execute.")
    arrow_compression: Literal["zstd", "lz4"] | None = Field(
        default=None,
        description="The compression algorithm to use for the Arrow response. Supported values are `zstd`, `lz4`, and `None`(meaning no compression).",
    )
    project_id: int | None = Field(
        default=None, description="If set, the query will only run over data associated with the given `project_id`."
    )
    version_id: int | None = Field(
        default=None,
        description="If set, will run the query against the given version of the data according to the changelog.",
    )


class StructType(BaseModel):
    type: Literal[TypeNames.STRUCT] = Field(default=TypeNames.STRUCT)
    columns: list[Annotated[PrimitiveColumn | StructColumn | ArrayColumn, Field(discriminator="type")]]


class TableArchive(BaseModel):
    pass


class TableLoadConfig(BaseModel):
    columns: list[Annotated[PrimitiveColumn | StructColumn | ArrayColumn, Field(discriminator="type")]] = Field(
        description="The column layout of the files. All files in the data load destined for this table must have this exact column layout. For CSV files, only the specified columns determine how data is loaded. The first row (header) in each uploaded file is skipped. For Parquet files, columns will be matched by name and are thus not order-sensitive. Therefore it is recommended to use Parquet files for more robust data loading."
    )
    format: FileFormat


class TableRename(BaseModel):
    new_reference: str = Field(
        description="A unique name to use for referring to the table, such as `temperature_assay`. User-created\ntables should always use this simple `table_name` format. Cradle-created table references have the form\n`catalog.schema.table_name`, for example `cradle.results.engineered_sequences`.\n\nReferences must be unique per workspace. One exception to the rule is that once a table is archived the\nrefence is available to be used again. References are tied to a data version, i.e. when running a query\nat a historic version, a reference will point to the table that held the reference at this data version."
    )


class TaskProducer(BaseModel):
    kind: Literal[ArtifactProducerKind.TASK] = Field(default=ArtifactProducerKind.TASK)
    task_id: int


class UserProducer(BaseModel):
    kind: Literal[ArtifactProducerKind.USER] = Field(default=ArtifactProducerKind.USER)


class ArrayColumn(ColumnBase, ArrayType):
    """A column representing an array of elements.

    Unlike the other column types it is not nullable but the array can be empty.
    This matches BigQuery's behavior and reduces ambiguity.
    """


class BaseTableCreate(TableCreateBase):
    kind: Literal[TableKind.TABLE] = Field(
        default=TableKind.TABLE,
        description="Whether to create a *table* or a *view*. Tables are created with the purpose of uploading data\n        to them, while views are created as queries on top of existing tables or views. Think of tables as physical\n        entities that store data, while views are virtual entities that represent a query result.\n        ",
    )
    columns: list[Annotated[PrimitiveColumn | StructColumn | ArrayColumn, Field(discriminator="type")]] = Field(
        description="Columns of the table"
    )


class ViewTableCreate(TableCreateBase):
    kind: Literal[TableKind.VIEW] = Field(
        default=TableKind.VIEW,
        description="Whether to create a *table* or a *view*. Tables are created with the purpose of uploading data\n        to them, while views are created as queries on top of existing tables or views. Think of tables as physical\n        entities that store data, while views are virtual entities that represent a query result.\n        ",
    )
    query: str = Field(
        description="SQL query for the view. It must be a single SELECT statement with a limited, but still very powerful, subset of SQL. Subqueries, common table expressions, joins, and conditional expressions with all commonly used SQL operators are supported. Contact your customer success representative if you need to use a specific SQL feature that is not supported. If you have a query use case that is not covered by the currently permitted SQL expressions, contact support to discuss potential extensions."
    )
    columns: list[Annotated[PrimitiveColumn | StructColumn | ArrayColumn, Field(discriminator="type")]] | None = Field(
        default=None,
        description="Columns of the view. The column schema must match the query's result schema. If unspecified, a schema is automatically inferred from the query and column metadata will be copied over from the existing schema if the column already existed.",
    )


class BaseTableResponse(TableResponseBase):
    kind: Literal[TableKind.TABLE] = Field(default=TableKind.TABLE)


class ViewTableResponse(TableResponseBase):
    kind: Literal[TableKind.VIEW] = Field(default=TableKind.VIEW)
    query: str = Field(description="The SQL query that defines the view.")


class BaseTableUpdate(_TableUpdateBase):
    kind: Literal[TableKind.TABLE] = Field(default=TableKind.TABLE)
    columns: list[Annotated[PrimitiveColumn | StructColumn | ArrayColumn, Field(discriminator="type")]] | None = Field(
        default=None, description="Columns of the table."
    )


class ViewTableUpdate(_TableUpdateBase):
    kind: Literal[TableKind.VIEW] = Field(default=TableKind.VIEW)
    query: str | None = Field(
        default=None,
        description="SQL query for the view. It must be a single SELECT statement and only a subset of SQL is allowed. If unspecified, no change is made to the view query.If you have a query use case that is not covered by the currently permitted SQL expressions, contact support to discuss potential extensions.",
    )
    columns: list[Annotated[PrimitiveColumn | StructColumn | ArrayColumn, Field(discriminator="type")]] | None = Field(
        default=None,
        description="Columns of the view. The column schema must match the query's result schema. If unspecified, a schema is automatically inferred from the query and column metadata will be copied over from the existing schema if the column already existed.",
    )


class PrimitiveColumn(ColumnBase, PrimitiveType):
    """A column representing a primitive type."""

    nullable: bool = Field(default=True)


class StructColumn(ColumnBase, StructType):
    """A column representing a struct of fields."""

    nullable: bool = Field(default=True)


class TableLoadConfigResponse(TableLoadConfig):
    table_id: int | None = Field(
        description="The table the rows were loaded into. This is the table addressed by the table reference for this configuration at the time of finalization."
    )


Column: TypeAlias = Annotated[PrimitiveColumn | StructColumn | ArrayColumn, Field(discriminator="type")]
ColumnType: TypeAlias = Annotated[PrimitiveType | StructType | ArrayType, Field(discriminator="type")]
TableResponse: TypeAlias = Annotated[BaseTableResponse | ViewTableResponse, Field(discriminator="kind")]



import contextlib
import json
import logging
import time
from datetime import UTC, datetime, timedelta
from typing import Annotated, Literal, Self
from urllib.parse import urlencode, urljoin

import httpx
import jwt
import keyring
import keyring.errors
from keyring.backends.chainer import ChainerBackend
from pydantic import BaseModel, Field, TypeAdapter, ValidationError
from typing_extensions import Generator

from cradle.sdk.exceptions import ClientError

logger = logging.getLogger(__name__)


class ClientAuthError(Exception):
    def __init__(self, message: str):
        self.message = message
        super().__init__(f"Authentication Error: {message}")


class ReauthenticationRequiredError(ClientAuthError):
    def __init__(self, logout_url):
        self.logout_url = logout_url
        super().__init__("Reauthentication required")


class DeviceAuthResponse(BaseModel):
    device_code: str
    user_code: str
    verification_uri: str
    verification_uri_complete: str
    expires_in: int
    interval: int


class AuthenticateSuccessResponse(BaseModel):
    access_token: str
    refresh_token: str
    organization_id: str | None = None
    authentication_method: str
    user: dict


class AuthenticateAuthenticationErrorResponse(BaseModel):
    error: str
    error_description: str


class UnknownErrorResponse(BaseModel):
    response: dict


AuthenticateErrorResponse = AuthenticateAuthenticationErrorResponse | UnknownErrorResponse


class UnauthorizedResponse(BaseModel):
    authorized: Literal[False] = False
    can_reauth: bool
    org_id: str | None


class AuthorizedResponse(BaseModel):
    authorized: Literal[True] = True


CheckAuthResponse = Annotated[UnauthorizedResponse | AuthorizedResponse, Field(discriminator="authorized")]
CheckAuthResponseAdapter = TypeAdapter(CheckAuthResponse)


class _DeviceAuthStrategy(BaseModel):
    client_id: str
    app_url: str
    auth_api_url: str


class DeviceAuth(httpx.Auth):
    def __init__(
        self,
        client_id: str,
        cradle_app_url: str,
        cradle_api_base_url: str,
        auth_base_url: str,
        workspace: str | None = None,
        use_keyring: bool = True,
        user_agent: str | None = None,
    ):
        self.client_id = client_id
        self.cradle_app_url = cradle_app_url
        self.use_keyring = use_keyring
        self._keyring_servicename: str = __name__
        self.workspace = workspace
        self.cradle_api_base_url = cradle_api_base_url
        self.token: AuthenticateSuccessResponse | None = None

        self.checked_for_reauth = False
        self.force_org_id = None

        self.auth_base_url = auth_base_url
        self.user_agent = user_agent

        # Try to load token from keyring on initialization
        if self.use_keyring:
            self._load_token_from_keyring()

    @classmethod
    def from_strategy(
        cls,
        client: httpx.Client,
        base_url: str,
        workspace: str | None,
        use_keyring: bool = True,
    ) -> Self:
        """Create device auth with parameters looked up in the API."""
        api_strategy_url = f"{base_url}/auth:apiStrategy"

        try:
            response = client.get(api_strategy_url)
            response.raise_for_status()
            strategy = _DeviceAuthStrategy.model_validate(response.json())
        except Exception as exc:
            raise RuntimeError(f"Failed to retreive auth strategy from {api_strategy_url}") from exc

        return cls(
            workspace=workspace,
            cradle_api_base_url=base_url,
            client_id=strategy.client_id,
            cradle_app_url=strategy.app_url,
            auth_base_url="https://auth.cradle.bio",
            use_keyring=use_keyring,
            user_agent=client.headers.get("User-Agent"),
        )

    @property
    def access_token(self) -> str | None:
        return self.token.access_token if self.token else None

    @property
    def refresh_token(self) -> str | None:
        return self.token.refresh_token if self.token else None

    @property
    def session_id(self) -> str | None:
        return self.jwt_payload.get("sid") if self.jwt_payload else None

    @property
    def authorized_workspace_id(self) -> str | None:
        return self.jwt_payload.get("urn:cradle:workspace_id") if self.jwt_payload else None

    @property
    def jwks_url(self) -> str:
        return f"{self.auth_base_url}/sso/jwks/{self.client_id}"

    @property
    def jwt_payload(self) -> dict | None:
        if self.token is None:
            return None

        # Per https://github.com/jpadilla/pyjwt/issues/939 - we should not be verifying the `iat` time here
        # as it could be slightly in the future.
        return jwt.decode(self.token.access_token, options={"verify_signature": False, "verify_iat": False})

    @property
    def expires_at(self) -> datetime | None:
        if self.jwt_payload is None:
            return None
        return datetime.fromtimestamp(self.jwt_payload["exp"], tz=UTC)

    @property
    def org_id(self) -> str | None:
        if self.jwt_payload is None:
            return None
        return self.jwt_payload.get("org_id")

    @property
    def logout_url(self) -> str | None:
        session_id = self.session_id
        if session_id is None:
            return None

        logout_path = "/user_management/sessions/logout"
        return_to = f"{self.cradle_app_url}/_/sdk/post-logout"
        query = urlencode({"session_id": session_id, "return_to": return_to})
        return urljoin(self.auth_base_url, logout_path) + f"?{query}"

    def _needs_refresh(self) -> bool:
        if self.access_token is None or self.expires_at is None:
            return False

        if self.force_org_id is not None and self.org_id != self.force_org_id:
            return True

        # access_token is valid until expires_at (which I think is currently issue time + 5 min)
        # We refresh the access_token 10 seconds before expiration to avoid race conditions
        return datetime.now(tz=UTC) > self.expires_at - timedelta(seconds=10)

    def _save_token_to_keyring(self, token: AuthenticateSuccessResponse) -> None:
        """Save token data to keyring."""
        if not self.use_keyring:
            return

        # The backends built into keyring are secure.
        # They will only be in the list of chainer backends if they are "viable"
        # If this list is empty, it likely means we have no secure keyrings available.
        has_keyring_backend = any(x.__class__.__module__.startswith("keyring.") for x in ChainerBackend.backends)
        # The backends built into keyrings.alt are insecure.
        has_alt_backend = any(x.__class__.__module__.startswith("keyrings.alt") for x in ChainerBackend.backends)

        if not has_keyring_backend and has_alt_backend:
            logger.warning(
                "Keyring is likely using a backend from `keyrings.alt`. This is an *insecure* way to save your authentication for a limited period of time."
            )

        # Store the token response, don't worry if saving to keyring fails - maybe this system doesn't support it
        #
        # Some systems don't support keyring, such as a headless linux machine (because there is no GUI to pop up the
        # "do you want to allow this application to access your keyring?" dialog). This can be a problem when people run
        # code on a remote machine.
        #
        # In those cases, a user can install `keyrings.alt` into their notebook or environment, and a new backend called
        # `keyrings.alt.file.PlaintextKeyring` will be available. Because this is insecure, we don't do it by default.
        # However, all a user needs to do is install `keyrings.alt` into their notebook or environment, and it will automatically
        # be added to the resolution order of keyring backends.
        #
        # If the system does not support any secure keyrings, and no alternate backends are installed, we will get a NoKeyringError.
        #
        # If there is a keyring that does not support set_password in the resolution order - such as keyrings.gauth.GooglePythonAuth
        # aka `keyrings.google-artifactregistry-auth`, we will get a NotImplementedError. If there is another keyring below it in the
        # order, like PlaintextKeyring that does, then the ChainerBackend will fall through to the PlaintextKeyring.
        try:
            keyring.set_password(self._keyring_servicename, self.client_id, token.model_dump_json())
        except (NotImplementedError, keyring.errors.NoKeyringError):
            logger.warning("""No keyring available. You will need to re-authenticate every time you run this code.

To save your authentication for a limited period of time in an *insecure* way, install python package `keyrings.alt` into your notebook or environment and re-run this code.
""")

    def _load_token_from_keyring(self) -> None:
        """Load token data from keyring."""
        if not self.use_keyring:
            return

        with contextlib.suppress(keyring.errors.KeyringError):
            token_json = keyring.get_password(self._keyring_servicename, self.client_id)
            if token_json is not None:
                self._update_token(AuthenticateSuccessResponse.model_validate(json.loads(token_json)))

    def _clear_token_from_keyring(self) -> None:
        """Clear token data from keyring."""
        if not self.use_keyring:
            return

        # Don't worry if deleting from keyring fails - maybe this system doesn't support it
        with contextlib.suppress(keyring.errors.KeyringError):
            keyring.delete_password(self._keyring_servicename, self.client_id)

    def _update_token(self, success_response: AuthenticateSuccessResponse) -> None:
        self.token = success_response
        # Save to keyring for persistence
        self._save_token_to_keyring(success_response)

    @staticmethod
    def parse_response(response: httpx.Response) -> AuthenticateSuccessResponse | AuthenticateErrorResponse:
        if response.is_success:
            return AuthenticateSuccessResponse.model_validate(response.json())
        try:
            return AuthenticateAuthenticationErrorResponse.model_validate(response.json())
        except ValidationError:
            return UnknownErrorResponse(response=response.json())

    def _authenticate(
        self, data: dict[str, str]
    ) -> Generator[httpx.Request, httpx.Response, AuthenticateSuccessResponse | AuthenticateErrorResponse]:
        response = yield httpx.Request(
            method="POST",
            url=f"{self.auth_base_url}/user_management/authenticate",
            headers={"Content-Type": "application/x-www-form-urlencoded", "User-Agent": self.user_agent or "unknown"},
            data={**data, "client_id": self.client_id},
        )
        response.read()

        return self.parse_response(response)

    def _refresh_token(self) -> Generator[httpx.Request, httpx.Response, None]:
        if self.refresh_token is None:
            raise ClientAuthError("No refresh token available")

        refresh_data = {
            "grant_type": "refresh_token",
            "refresh_token": self.refresh_token,
        }
        if self.force_org_id is not None:
            refresh_data["organization_id"] = self.force_org_id

        response = yield from self._authenticate(refresh_data)

        match response:
            case AuthenticateAuthenticationErrorResponse(error="invalid_grant"):
                # This might mean the refresh token has expired, so let's clear keyring and re-run the device flow
                self._clear_token_from_keyring()
                self.token = None
                yield from self._start_device_flow()
            case AuthenticateAuthenticationErrorResponse(error="sso_required"):
                # This means the token cannot be refreshed for a different org that requires SSO,
                # so let's clear keyring and re-run the device flow
                self._clear_token_from_keyring()
                self.token = None
                yield from self._start_device_flow()
            case AuthenticateAuthenticationErrorResponse(error="mfa_enrollment"):
                # It might not be possible to refresh a token to a different organization if the new
                # organization has a different security policy in place, e.g. regarding MFA.
                # In this case we need to log out and re-authenticate to the new workspace.
                # Unfortunately this requires the user to manually click a logout link.
                logout_url = self.logout_url
                if logout_url is None:
                    raise ClientAuthError("No session ID available to log out.")

                print(
                    "Access to the requested workspace requires re-authentication. "
                    "Please log out under the following link and then retry the request.\n\n"
                    f"{logout_url}"
                )

                raise ReauthenticationRequiredError(logout_url)
            case AuthenticateAuthenticationErrorResponse():
                raise ClientAuthError(response.error_description)
            case UnknownErrorResponse():
                raise ClientAuthError(json.dumps(response.response))
            case AuthenticateSuccessResponse():
                self._update_token(response)

    def _poll_for_tokens(
        self, device_code: str, expires_in: int = 300, interval: int = 5
    ) -> Generator[httpx.Request, httpx.Response, AuthenticateSuccessResponse]:
        """Poll for authentication tokens using device code flow.

        Args:
            device_code: The device code from the initial auth request
            expires_in: Timeout in seconds (default 300)
            interval: Polling interval in seconds (default 5)

        Returns:
            AuthenticateSuccessResponse: access and refresh tokens

        Raises:
            Exception: If polling for access/refresh tokens fails or times out
        """
        start_time = time.monotonic()

        while True:
            # Check timeout
            if time.monotonic() - start_time > expires_in:
                raise ClientAuthError("Authentication timed out")

            try:
                refresh_data = {
                    "device_code": device_code,
                    "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                    "organization_id": self.force_org_id,
                }
                response = yield from self._authenticate(refresh_data)

            except httpx.TimeoutException as e:
                raise ClientAuthError("Authentication timed out") from e

            match response:
                case AuthenticateSuccessResponse():
                    return response
                case AuthenticateAuthenticationErrorResponse(error="authorization_pending"):
                    time.sleep(interval)
                case AuthenticateAuthenticationErrorResponse(error="slow_down"):
                    interval += 1
                case AuthenticateAuthenticationErrorResponse():
                    raise ClientAuthError(response.error_description)
                case UnknownErrorResponse():
                    raise ClientAuthError(json.dumps(response.response))

    def _start_device_flow(self) -> Generator[httpx.Request, httpx.Response, None]:
        device_auth_http_response: httpx.Response = yield httpx.Request(
            method="POST",
            url=f"{self.auth_base_url}/user_management/authorize/device",
            json={
                "client_id": self.client_id,
            },
            headers={"user-agent": self.user_agent or "unknown"},
        )
        device_auth_http_response.read()
        device_auth_http_response.raise_for_status()
        device_auth_response = DeviceAuthResponse.model_validate(device_auth_http_response.json())

        print(f"""
Please click the following URL to complete authentication

{device_auth_response.verification_uri_complete}

and verify that the code matches: {device_auth_response.user_code}
""")

        # Poll for tokens using the device code
        token_data = yield from self._poll_for_tokens(
            device_code=device_auth_response.device_code,
            expires_in=device_auth_response.expires_in,
            interval=device_auth_response.interval,
        )
        self._update_token(token_data)

    def _check_auth(self) -> Generator[httpx.Request, httpx.Response, None | CheckAuthResponse]:
        if self.cradle_api_base_url is None or self.workspace is None or self.access_token is None:
            return None

        needs_reauth_response: httpx.Response = yield httpx.Request(
            method="GET",
            url=f"{self.cradle_api_base_url}/auth:checkAuth",
            params={"workspace": self.workspace},
            headers={"Authorization": f"Bearer ca_{self.access_token}", "User-Agent": self.user_agent or "unknown"},
        )

        needs_reauth_response.read()
        needs_reauth_response.raise_for_status()
        return CheckAuthResponseAdapter.validate_python(needs_reauth_response.json())

    def auth_flow(self, request):
        if not self.access_token:
            yield from self._start_device_flow()

        # If we loaded the token from keyring, the access_token is almost certainly expired,
        # so we need to refresh it *before* hitting the checkAuth endpoint, which is authenticated
        # otherwise we get a 403.
        if self._needs_refresh():
            yield from self._refresh_token()

        # Only check for reauth once - the requested workspace that this client is constructed with
        # will not change, so we won't need to check if we need to reauth to a different workspace/org
        # more than once
        if not self.checked_for_reauth:
            check_auth_response = yield from self._check_auth()
            # can_reauth=True means that we are a member of the workspace we want to access, but need to reauthorize
            # our access_token to the correct org
            if check_auth_response and check_auth_response.authorized == False and check_auth_response.can_reauth:  # noqa: E712
                # This will force _refresh_token to request access to the correct org
                self.force_org_id = check_auth_response.org_id

            self.checked_for_reauth = True

        # After checking if we need to reauth, this will fire if we need to re-auth the token to a different org
        # from the one it was issued for.
        if self._needs_refresh():
            yield from self._refresh_token()

        request.headers["Authorization"] = f"Bearer ca_{self.access_token}"
        yield request

    def suggest_logout(self, workspace: str) -> None:  # noqa: ARG002 - might use workspace in the future
        """Print logout URL."""
        logout_url = self.logout_url
        if logout_url is None:
            raise ClientAuthError("No session ID available to log out.")

        # Don't clear token in case this is caused by fetching a workspace immediately after
        # creating it.
        #
        # Don't clear keyring in case the user decides to recreate the client with a different
        # workspace name.

        # Reset this check in case we just created a workspace and now need to auth to it
        self.checked_for_reauth = False

        message = (
            f'Authorized workspace "{self.authorized_workspace_id}" does not match requested workspace "{self.workspace}"\n'
            f"If you believe this is an error, you can log out of this session at {logout_url} and then try the request again."
        )
        raise ClientError(403, message, [])



import csv as csv_std
import tempfile
import time
from abc import ABC, abstractmethod
from collections.abc import Callable
from datetime import UTC, datetime, timedelta
from pathlib import Path
from typing import Self

import openpyxl
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.csv as csv
import pyarrow.parquet as pq
from typing_extensions import TypeVar

from cradle.sdk.client import DataLoadClient
from cradle.sdk.types.common import ContextProject, ContextRound
from cradle.sdk.types.data import (
    ArrayColumn,
    ArrayType,
    Column,
    ColumnType,
    DataLoadCreate,
    DataLoadResponse,
    DataLoadState,
    FileFormat,
    PrimitiveColumn,
    PrimitiveType,
    StructColumn,
    StructType,
    TableLoadConfig,
    TypeNames,
)


class Transform(ABC):
    """Base class for all transformations applied to UploadFile objects."""

    @abstractmethod
    def apply(self, data: pa.Table) -> pa.Table:
        """Apply the transformation to an UploadFile and return a new UploadFile."""
        ...

    def __call__(self, data: pa.Table) -> pa.Table:
        return self.apply(data)


class File:
    def __init__(self, data: pa.Table, source: Path | None, relative_path: Path | None = None):
        self.data = data
        self.source = source
        self.relative_path = relative_path
        self.columns: list[Column] = []

        try:
            for field in data.schema:
                self.columns.append(_convert_column(field.name, field.type))
        except _NullTypeError as e:
            raise ValueError(f"""
                {e}

                For CSV input files, provide an explicit column type (e.g. `convert_options=pyarrow.csv.ConvertOptions(column_types={{"c1": pa.string()}})`)
                or omit the affected column (via `ConvertOptions(include_columns=[...])`).
            """) from e


class FileSet:
    def __init__(self, files: list[File]):
        if len(files) == 0:
            raise ValueError("file list must not be empty")

        for f in files[1:]:
            if f.columns != files[0].columns:
                raise ValueError(
                    f"all files must have the same schema, found {f.columns} but expected {files[0].columns}"
                )

        self.files = files

    @property
    def columns(self) -> list[Column]:
        return self.files[0].columns

    def merge(self, other: "FileSet") -> "FileSet":
        if self.columns != other.columns:
            raise ValueError(f"FileSets have incompatible columns {self.columns} and {other.columns}")
        return FileSet(self.files + other.files)

    def transform(self, *transforms: Transform) -> "FileSet":
        """Apply all transformations to all files.

        Use this primarily to make the data in the files fit the schema of the target table.
        Avoid transformations that prepare data for a specific task. Those use cases should
        instead be deferred to views over base tables in the platform.
        """

        # Iterate over transforms in the outer loop so we know for each transform
        # whether it succeeds for all files before proceeding. This should make
        # debugging overall easier.
        files = self.files[:]
        for t in transforms:
            files = [File(t(f.data), f.source, f.relative_path) for f in files]
        return FileSet(files)

    @classmethod
    def _from_files(
        cls,
        open_fn: Callable[[Path], pa.Table],
        directory: str | Path,
        pattern: str | list[str],
    ) -> Self:
        if isinstance(directory, str):
            directory = Path(directory)
        if isinstance(pattern, str):
            pattern = [pattern]

        files = []
        for p in pattern:
            for f in directory.glob(p):
                relative_path = f.relative_to(directory)
                files.append(File(data=open_fn(f), source=f, relative_path=relative_path))
        return cls(files)

    @classmethod
    def from_parquet(
        cls,
        directory: Path | str,
        pattern: str | list[str],
        **parquet_options,
    ) -> Self:
        """Create a `FileSet` from Parquet source files.

        Args:
            directory: The base directory relative to which the file path pattern is matched.
            pattern: The glob pattern or list of patterns to match the Parquet files. The resulting paths
                relative to `directory` will be submitted as metadata with the uploaded files and be visible
                in the platform.
            **parquet_options: Additional options to pass to `pyarrow.parquet.read_table` to control how the
                Parquet files are read.
        """

        def _open(file: Path) -> pa.Table:
            return pq.read_table(file, **parquet_options)

        return cls._from_files(_open, directory, pattern)

    @classmethod
    def from_csv(
        cls,
        directory: Path | str,
        pattern: str | list[str],
        **csv_options,
    ) -> Self:
        """Create a `FileSet` from CSV source files.

        Args:
            directory: The base directory relative to which the file path pattern is matched.
            pattern: The glob pattern or list of patterns to match the CSV files. The resulting paths
                relative to `directory` will be submitted as metadata with the uploaded files and be visible
                in the platform.
            **csv_options: Additional options to pass to `pyarrow.csv.read_csv` to control how the
                CSV files are read.
        """

        def _open(file: Path) -> pa.Table:
            return csv.read_csv(file, **csv_options)

        return cls._from_files(_open, directory, pattern)

    @classmethod
    def from_excel(
        cls,
        directory: Path | str,
        pattern: str | list[str],
        sheet: str | int,
        **csv_options,
    ) -> Self:
        """Create a `FileSet` from Excel files (xlsx or xls).

        Excel files are first converted to CSV as-is and then read as CSV files like in `from_csv`.

        Args:
            directory: The base directory relative to which the file path pattern is matched.
            pattern: The glob pattern or list of patterns to match the Excel files. The resulting paths
                relative to `directory` will be submitted as metadata with the uploaded files and be visible
                in the platform.
            sheet: The sheet name or index to read from the Excel file. If multiple sheets should be read,
                call `from_excel` individually for each sheet.
            **csv_options: Additional options to pass to `pyarrow.csv.read_csv` to control how the
                converted CSV files are read.
        """

        def _open(file: Path) -> pa.Table:
            with tempfile.NamedTemporaryFile(suffix=".csv") as temp_csv:
                tmp_csv = Path(temp_csv.name)

                workbook = openpyxl.load_workbook(file, read_only=True, data_only=True)
                if isinstance(sheet, str):
                    if sheet not in workbook.sheetnames:
                        raise ValueError(f"Sheet '{sheet}' not found in {file}")
                    data = workbook[sheet]
                else:
                    if sheet >= len(workbook.sheetnames):
                        raise ValueError(f"Sheet index {sheet} is out of range for {file}")
                    data = workbook[workbook.sheetnames[sheet]]

                with tmp_csv.open("w", encoding="utf-8", newline="") as f:
                    writer = csv_std.writer(f)
                    for row in data.rows:
                        writer.writerow([cell.value for cell in row])

                return csv.read_csv(tmp_csv, **csv_options)

        return cls._from_files(_open, directory, pattern)


class Uploader:
    def __init__(self, client: DataLoadClient, context: ContextProject | ContextRound):
        """Initialize the Uploader with a Client instance.

        Args:
            client: An authenticated Client instance
            context: The context of the upload
        """
        self._load_id: int | None = None
        self._client = client
        self._context = context
        self._tables: dict[str, FileSet] = {}

    def add_files(self, table_name: str, files: FileSet) -> None:
        """Add files to the upload for the specified table.

        All files for a given table must have the same schema, which must be compatible
        with the schema of the target table. It is compatible if all required columns exist
        and have the right type.

        Args:
            table_name: The name of the table the files will be ingested into.
            files: The files to upload.
        """
        existing = self._tables.get(table_name)
        if existing is not None:
            files = existing.merge(files)
        self._tables[table_name] = files

    def load(self) -> DataLoadResponse:
        """Upload all files for this upload.

        For each added `UploadFile` this will upload the data in Parquet format to be ingested
        into the target table. Additionally, the source, if set, will be uploaded for archival purposes.

        The upload will be left in pending state until finalize is called.

        Returns:
            UploadResponse: The upload state retrieved from the server after all uploads completed.
        """
        if self._load_id is not None:
            raise ValueError("Upload already started")

        load = self._client.create(
            DataLoadCreate(
                context=self._context,
                tables={
                    name: TableLoadConfig(columns=files.columns, format=FileFormat.PARQUET)
                    for name, files in self._tables.items()
                },
            )
        )
        self._load_id = load.id

        with tempfile.TemporaryDirectory(prefix="cradle-upload-") as temp_dir:
            for ref, t in self._tables.items():
                base_dir = Path(temp_dir) / ref
                base_dir.mkdir(exist_ok=False, parents=True)

                for i, file in enumerate(t.files):
                    description = None
                    source_file_id = None
                    if file.source is not None:
                        print(f"Uploading source file {file.source}...")
                        resp = self._client.upload_file(
                            load_id=self._load_id,
                            file=file.source,
                            filepath=file.relative_path,
                            table_reference=None,
                        )
                        description = f"Source file: {file.source} ({resp.id})"
                        source_file_id = resp.id

                    path = base_dir / f"{i}.parquet"
                    print(f"Uploading file {path}...")
                    pq.write_table(file.data, path)
                    self._client.upload_file(
                        load_id=self._load_id,
                        file=path,
                        filepath=(
                            file.relative_path.with_suffix(".parquet") if file.relative_path is not None else None
                        ),
                        table_reference=ref,
                        description=description,
                        source_file_id=source_file_id,
                    )

        return self._client.get(load_id=self._load_id)

    def finalize(self, wait: bool = True, timeout: float = 60) -> DataLoadResponse:
        """Finalize the upload. Calls `upload()` if it has not already been called.

        Args:
            wait: Whether to wait for the upload to complete.
            timeout: Timeout in seconds for waiting for the upload to complete.

        Returns:
            UploadResponse: The most recent upload state retrieved from the server.
        """
        if self._load_id is None:
            load_id = self.load().id
        else:
            load_id = self._load_id

        u = self._client.finalize(load_id=load_id)
        if not wait:
            return u

        def _fn():
            u = self._client.get(load_id=load_id)
            if u.state == DataLoadState.FAILED:
                raise RuntimeError(f"Upload failed: {', '.join(u.errors)}")
            return u.state == DataLoadState.COMPLETED, u

        return _wait_for_condition(_fn, datetime.now(tz=UTC) + timedelta(seconds=timeout))


class RenameColumns(Transform):
    """Transform that renames columns in the dataframe."""

    def __init__(self, column_map: dict[str, str]):
        """Args:
        column_map: A dictionary mapping old column names to new column names
        """
        self.column_map = column_map

    def apply(self, data: pa.Table) -> pa.Table:
        names = [self.column_map.get(name, name) for name in data.column_names]
        return data.rename_columns(names)


class FilterRows(Transform):
    """Filter rows in the dataframe based on a provided condition.

    Example usage: FilterRows(lambda t: pa.compute.greater(t.column("a"), 42))
    """

    def __init__(self, condition: Callable[[pa.Table], pa.Array]):
        """Args:
        condition: A function that computes a boolean array over all rows in the provided tables.
            Rows for which the result index is False, will be dropped from the resulting table.
        """
        self._condition = condition

    def apply(self, data: pa.Table) -> pa.Table:
        return data.filter(self._condition(data))


class DropNullRows(FilterRows):
    """Transform that removes rows with null values in the specified columns."""

    def __init__(self, columns: str | list[str] | None = None):
        """Args:
        columns: Column or list of column names to check for nulls. If None, checks all columns.
        """
        if isinstance(columns, str):
            columns = [columns]
        self._columns = columns

    def apply(self, data: pa.Table) -> pa.Table:
        columns = self._columns
        if columns is None:
            columns = data.column_names
        mask = pa.array([True] * len(data))

        for col in columns:
            if col not in data.column_names:
                raise ValueError(f"Column '{col}' does not exist in table")
            mask = pc.and_(mask, pc.is_valid(data[col]))  # type: ignore[reportAttributeAccessIssue] seem to actually be missing in venv, but invoking them works ???

        return data.filter(mask)


T = TypeVar("T")


def _wait_for_condition(cond_fn: Callable[[], tuple[bool, T]], deadline: datetime, interval: float = 1) -> T:
    """Wait for the condition function to return True or the deadline is exceeded. It will be executed at least once."""

    while True:
        done, result = cond_fn()
        if done:
            return result
        if datetime.now(tz=UTC) > deadline:
            raise TimeoutError("Deadline exceeded")
        time.sleep(interval)


class _NullTypeError(ValueError): ...


def _convert_type(type_: pa.DataType) -> ColumnType:
    if pa.types.is_integer(type_):
        return PrimitiveType(type=TypeNames.INT64)
    if pa.types.is_floating(type_):
        return PrimitiveType(type=TypeNames.FLOAT64)
    if pa.types.is_boolean(type_):
        return PrimitiveType(type=TypeNames.BOOL)
    if pa.types.is_string(type_):
        return PrimitiveType(type=TypeNames.STRING)
    if pa.types.is_struct(type_):
        return StructType(columns=[_convert_column(f.name, f.type) for f in type_.fields])
    if pa.types.is_list(type_):
        return ArrayType(item_type=_convert_type(type_.value_type))
    if pa.types.is_null(type_):
        raise _NullTypeError(f"Unsupported PyArrow type {type_}")
    else:
        raise ValueError(f"Unsupported PyArrow type {type_}")


def _convert_column(name: str, type_: pa.DataType) -> Column:
    if pa.types.is_integer(type_):
        return PrimitiveColumn(name=name, type=TypeNames.INT64, nullable=True)
    if pa.types.is_floating(type_):
        return PrimitiveColumn(name=name, type=TypeNames.FLOAT64, nullable=True)
    if pa.types.is_boolean(type_):
        return PrimitiveColumn(name=name, type=TypeNames.BOOL, nullable=True)
    if pa.types.is_string(type_):
        return PrimitiveColumn(name=name, type=TypeNames.STRING, nullable=True)
    if pa.types.is_struct(type_):
        columns = [_convert_column(f.name, f.type) for f in type_.fields]
        return StructColumn(name=name, columns=columns, nullable=True)
    if pa.types.is_list(type_):
        return ArrayColumn(name=name, item_type=_convert_type(type_.value_type))
    if pa.types.is_null(type_):
        raise _NullTypeError(f"Unsupported PyArrow type {type_} for column {name}")
    else:
        raise ValueError(f"Unsupported PyArrow type {type_} for column {name}")



import copy
import difflib
import logging
import time
from collections.abc import Iterator
from importlib.metadata import version as pkg_version
from pathlib import Path
from typing import Any, BinaryIO, NoReturn, TypeVar
from urllib.parse import urlparse

import httpx
import jwt
import pyarrow as pa
from pydantic import BaseModel, TypeAdapter
from typing_extensions import Generator, override

from cradle.sdk.auth.device import DeviceAuth
from cradle.sdk.exceptions import ClientError as ClientError  # for re-export
from cradle.sdk.types.common import (
    ContextProject,
    ContextRound,
    ContextWorkspace,
    ListOptions,
)
from cradle.sdk.types.data import (
    AddTableRequest,
    ArtifactResponse,
    BaseTableCreate,
    BaseTableUpdate,
    DataLoadCreate,
    DataLoadResponse,
    DataVersionResponse,
    FileUploadResponse,
    ListArtifactResponse,
    ListDataLoadResponse,
    ListDataVersionResponse,
    ListTableResponse,
    QueryDataRequest,
    TableArchive,
    TableRename,
    TableResponse,
    ViewTableCreate,
    ViewTableUpdate,
)
from cradle.sdk.types.task import (
    ListTaskResponse,
    TaskCreate,
    TaskResponse,
    TaskState,
)
from cradle.sdk.types.workspace import (
    ListProjectResponse,
    ListRoundResponse,
    ProjectCreate,
    ProjectResponse,
    RoundCreate,
    RoundResponse,
    WorkspaceResponse,
)

logger = logging.getLogger(__name__)

T = TypeVar("T", bound=BaseModel)


class HttpClient:
    def __init__(
        self,
        prefix: str,
        base_url: str,
        auth: httpx.Auth | None = None,
        timeout: float = 60.0,
        user_agent: str | None = None,
        workspace: str | None = None,
        use_keyring: bool = True,
    ):
        base_url = base_url.rstrip("/")
        user_agent = user_agent or f"cradle-sdk-python/{pkg_version('cradle-sdk')}"
        headers = {
            "Accept": "application/json",
            "User-Agent": user_agent,
        }
        client = httpx.Client(auth=auth, headers=headers, timeout=timeout, follow_redirects=True)
        if auth is None:
            client.auth = DeviceAuth.from_strategy(
                client=client, workspace=workspace, base_url=base_url, use_keyring=use_keyring
            )

        self.http_client = client
        self.prefix = prefix.strip("/")
        self.base_url = base_url.rstrip("/")
        self.workspace = workspace

    def with_prefix(self, prefix: str) -> "HttpClient":
        c = copy.copy(self)
        c.prefix = f"{self.prefix}/{prefix.strip('/')}".strip("/")
        return c

    def url(self, path: str) -> str:
        """Generate a URL for a given path.

        This follows slightly weird rules, specific to how our API is designed.

        The overall goal here is to make the URL generation loose with what it accepts, to never
        be confused by whether or not there are unexpected slashes in any part of the url. It's a very
        "do what I mean" approach.

        - "path" is never interpreted as an absolute path, it is always relative to the prefix
        - if the path starts with ":", it is a verb endpoint and should be appended to the url + prefix, not assumed to be a subpath
        - everything else is assumed to be a subpath and should be appended to the url + prefix
        """

        combined_path = f"{self.prefix}{path}"
        if not path.startswith(":") and not path.startswith("/"):
            combined_path = f"{self.prefix}/{path}"

        # This special case is needed in case we have a ":verb" endpoint at the root of the API
        # like "https://cradle.bio/:list" - otherwise the logic below would generate
        # "https://cradle.bio:list" which is wrong.
        parsed_url = urlparse(self.base_url)
        base_url_has_path = bool(parsed_url.path and parsed_url.path not in {"", "/"})
        if combined_path.startswith(":") and base_url_has_path:
            return f"{self.base_url}{combined_path}"

        return f"{self.base_url}/{combined_path.lstrip('/')}"

    def request(
        self,
        method: str,
        path: str,
        json: dict[str, Any] | BaseModel | None = None,
        params: dict[str, Any] | None = None,
        **kwargs,
    ) -> httpx.Response:
        if params is not None:
            # HTTPX will send "...&foo=&...", which FastAPI would interpret as an empty string rather than None.
            kwargs["params"] = {k: v for k, v in params.items() if v is not None}
        if isinstance(json, BaseModel):
            json = json.model_dump(mode="json", by_alias=True)
        if json is not None:
            kwargs["json"] = json

        return self.http_client.request(method, self.url(path), **kwargs)

    def get(self, path: str, response_type: type[T] | Any | None, params: dict[str, Any] | None = None, **kwargs) -> T:
        return self._handle_response(self.request("GET", path, params=params, **kwargs), response_type)

    def post(self, path: str, response_type: type[T] | Any | None, params: dict[str, Any] | None = None, **kwargs) -> T:
        return self._handle_response(self.request("POST", path, params=params, **kwargs), response_type)

    def _handle_response(
        self, response: httpx.Response, response_type: type[T] | Iterator[str] | Iterator[bytes] | None
    ) -> Any:
        if not response.is_success:
            self.handle_error_response(response)
        if response_type is None:
            return None

        ct = response.headers.get("content-type")
        if ct is None:
            raise ValueError("content type header missing")

        ct = ct.split(";")[0].strip()
        if ct == "application/json":
            return TypeAdapter(response_type).validate_python(response.json())
        if ct == "application/x-ndjson" and response_type == Iterator[str]:
            return response.iter_lines()
        if ct == "application/octet-stream" and response_type == Iterator[bytes]:
            return response.iter_bytes()
        raise ValueError(f"Unsupported content type {ct}")

    def handle_error_response(self, response: httpx.Response) -> NoReturn:
        auth = self.http_client.auth
        if (
            response.status_code == 403
            and self.workspace is not None
            and isinstance(auth, DeviceAuth)
            and auth.authorized_workspace_id != self.workspace
        ):
            auth.suggest_logout(self.workspace)
        try:
            body = response.json()
            error_msg = body.get("detail", response.text)
            if (error_id := body.get("error_id")) is not None:
                error_msg = f"{error_msg} (error ID: {error_id})"
            errors = body.get("errors", [])
        except Exception:  # noqa: BLE001
            error_msg = response.text
            errors = []
        raise ClientError(response.status_code, error_msg, errors=errors)


class DisabledAuth(httpx.Auth):
    """Auth implementation to use when authentication should be disabled"""

    def __init__(self, workspace: str | None = None):
        self._workspace_name = workspace

    @override
    def auth_flow(self, request: httpx.Request) -> Generator[httpx.Request, httpx.Response, None]:
        claims = {
            "urn:cradle:email": "john-doe@cradle.bio",
            "urn:cradle:user_id": "john-doe",
            "urn:cradle:full_name": "John Doe",
            "urn:cradle:workspace_id": self._workspace_name,
        }

        token = jwt.encode(claims, key="none")

        request.headers["Authorization"] = f"Bearer ca_{token}"
        yield request


API_URL = "https://api.cradle.bio"


class Client:
    """The workspace API provides functionality for retrieving information about your workspace.

    A workspace is the home for all data, projects and tasks for an organization. By
    design no information can be shared across workspaces, they are the top-level
    concept in the Cradle API.

    Workspace creation and management is only available to Cradle administrators.
    """

    def __init__(
        self,
        workspace: str,
        auth: httpx.Auth | None = None,
        base_url: str = API_URL,
        timeout: int = 60,
        user_agent: str | None = None,
        use_keyring: bool = True,
    ):
        self._base_client = HttpClient(
            prefix="/",
            base_url=base_url,
            auth=auth,
            timeout=timeout,
            user_agent=user_agent,
            workspace=workspace,
            use_keyring=use_keyring,
        )
        self._client = self._base_client.with_prefix(f"/v2/workspace/{workspace}/")
        self._workspace = workspace

    def workspace(self) -> WorkspaceResponse:
        return self._base_client.get("/v2/workspace:get", WorkspaceResponse, dict(name=self._workspace))

    @property
    def project(self) -> "ProjectClient":
        return ProjectClient(self._client)

    @property
    def round(self) -> "RoundClient":
        return RoundClient(self._client)

    @property
    def data(self) -> "DataClient":
        return DataClient(self._client)

    @property
    def task(self) -> "TaskClient":
        return TaskClient(self._client)


class ProjectClient:
    """The projects API provides functionality for managing and interacting with
    projects in a workspace.

    Projects organize tasks in a workspace and represent the work involved in
    optimizing a specific protein or achieving a specific goal. Data is currently
    isolated to a specific project, but in the future we will support cross project
    data sharing within a workspace.
    """

    def __init__(self, client: HttpClient):
        self._client = client.with_prefix("/project")

    def get(self, project_id: int) -> ProjectResponse:
        """Get a project by `ID`."""
        return self._client.get(":get", ProjectResponse, dict(id=project_id))

    def list(self) -> Generator[ProjectResponse, None, None]:
        """List all projects in the workspace."""
        opts = ListOptions()
        while True:
            resp = self._client.get(":list", ListProjectResponse, params=opts.model_dump())
            yield from resp.items
            if resp.cursor is None:
                return
            opts.cursor = resp.cursor

    def get_by_name(self, name: str) -> ProjectResponse:
        """Get a project by name. Raises ValueError if not found."""
        for project in self.list():
            if project.name == name:
                return project
        raise ValueError(f"Project with name '{name}' not found")

    def create(self, project: ProjectCreate) -> ProjectResponse:
        """Create a new project in the workspace."""
        return self._client.post(":create", ProjectResponse, json=project)

    def update(self, project_id: int, project: ProjectCreate) -> ProjectResponse:
        """Update a project's information."""
        return self._client.post(":update", ProjectResponse, dict(id=project_id), json=project)

    def archive(self, project_id: int) -> ProjectResponse:
        """Archive a project.

        The data from the project will still be available.
        It will not be possible to create new tasks, data loads, etc. in the project.
        """
        return self._client.post(":archive", ProjectResponse, dict(id=project_id))

    def unarchive(self, project_id: int) -> ProjectResponse:
        """Unarchive a previously archived project.

        Returns an HTTP 422 `Unprocessable Entity` error if the project is not archived.
        """
        return self._client.post(":unarchive", ProjectResponse, dict(id=project_id))


class RoundClient:
    """The rounds API provides functionality for managing and interacting with rounds within a project.

    Rounds are a way to organize the experiments within a project. Commonly, a round represents all
    in-vitro experiments related to a specific sub-goal or timeframe of a project.
    While rounds are typically ordered sequentially, multiple rounds can exist in parallel (for example, when
    testing multiple design methods in parallel).

    Tasks, data loads, and reports can be (but don't have to be!) assigned to a round.
    """

    def __init__(self, client: HttpClient):
        self._client = client.with_prefix("/round")

    def get(self, round_id: int) -> RoundResponse:
        """Get a round by ID."""
        return self._client.get(":get", RoundResponse, dict(id=round_id))

    def list(self, project_id: int | None = None) -> Generator[RoundResponse, None, None]:
        """List all rounds in the project."""
        params = {}
        if project_id is not None:
            params["project_id"] = project_id

        opts = ListOptions()
        while True:
            resp = self._client.get(
                ":list",
                ListRoundResponse,
                params={**opts.model_dump(), **params},
            )
            yield from resp.items
            if resp.cursor is None:
                return
            opts.cursor = resp.cursor

    def get_by_name(self, project_id: int, name: str) -> RoundResponse:
        """Get a round by name within a specific project. Raises ValueError if not found."""
        for round_ in self.list(project_id=project_id):
            if round_.name == name:
                return round_
        raise ValueError(f"Round with name '{name}' not found in project {project_id}")

    def create(self, round_: RoundCreate) -> RoundResponse:
        """Create a new round in a project."""
        return self._client.post(":create", RoundResponse, json=round_)

    def update(self, round_id: int, round_: RoundCreate) -> RoundResponse:
        """Update a round's name and description."""
        return self._client.post(":update", RoundResponse, dict(id=round_id), json=round_)

    def archive(self, round_id: int) -> RoundResponse:
        """Archive a round.

        The data from the round will still be available.
        It will not be possible to create new tasks, data loads, etc. in the round.
        """
        return self._client.post(":archive", RoundResponse, dict(id=round_id))

    def unarchive(self, round_id: int) -> RoundResponse:
        """Unarchive a previously archived round.

        Returns an HTTP 422 `Unprocessable Entity` error if the round is not archived.
        """
        return self._client.post(":unarchive", RoundResponse, dict(id=round_id))


class DataClient:
    """The data API provides functionality for managing and interacting with data in a workspace.

    Data comes in the form of *tables* and *artifacts*.
    *Tables* are used to store row-based data, such as measurements, sequences, and other structured data.
    *Artifacts* are used to store the results of tasks, such as machine learning models, predictions, and other outputs.

    While tables can be created and uploaded by users, artifacts are only produced by tasks and cannot be uploaded directly.

    Artifacts and table rows (via their data load source) have a context specified. This context refers to the
    origin of the data and is used in other parts of the API to determine data visibility, e.g. to
    isolate data between projects.

    See the [Artifacts](#tag/Artifacts) API endpoints for more details on artifact data.

    See the [Tables](#tag/Tables) API endpoints for more details on tabular data.
    """

    def __init__(self, client: HttpClient):
        self._client = client.with_prefix("/data")

    @property
    def artifact(self) -> "ArtifactClient":
        return ArtifactClient(self._client)

    @property
    def table(self) -> "TableClient":
        return TableClient(self._client)

    @property
    def load(self) -> "DataLoadClient":
        return DataLoadClient(self._client)

    def list_versions(
        self,
        project_id: int | None = None,
        table_id: int | None = None,
    ) -> Generator[DataVersionResponse, None, None]:
        params = {}
        if project_id is not None:
            params["project_id"] = project_id
        if table_id is not None:
            params["table_id"] = table_id

        opts = ListOptions()
        while True:
            resp = self._client.get("/version:list", ListDataVersionResponse, params={**params, **opts.model_dump()})
            yield from resp.items
            if resp.cursor is None:
                return
            opts.cursor = resp.cursor


class ArtifactClient:
    """Artifacts are produced by tasks and can be referenced as inputs to other tasks. They can generally
    not be uploaded and downloaded directly. They have a well-defined type that determines for which
    purpose they can be used.
    """

    def __init__(self, client: HttpClient):
        self._client = client.with_prefix("/artifact")

    def get(self, artifact_id: int) -> ArtifactResponse:
        """Get information about an artifact."""
        return self._client.get(":get", ArtifactResponse, dict(id=artifact_id))

    def list(
        self,
        project_id: int | None = None,
        round_id: int | None = None,
    ) -> Generator[ArtifactResponse, None, None]:
        """List artifacts in the workspace."""
        opts = ListOptions()

        params = {}
        if project_id is not None:
            params["project_id"] = project_id
        if round_id is not None:
            params["round_id"] = round_id

        while True:
            resp = self._client.get(
                ":list",
                ListArtifactResponse,
                params={**opts.model_dump(), **params},
            )
            yield from resp.items
            if resp.cursor is None:
                return
            opts.cursor = resp.cursor


class _IteratorReader:
    def __init__(self, iterator: Iterator[bytes]):
        self.iterator = iterator
        self.buffer = b""
        self._closed = False

    @property
    def closed(self) -> bool:
        return self._closed

    def close(self) -> None:
        self._closed = True
        self._buffer = b""

    def read(self, size=-1):
        try:
            while size < 0 or len(self.buffer) < size:
                self.buffer += next(self.iterator)
        except StopIteration:
            pass
        if size < 0:
            out, self.buffer = self.buffer, b""
        else:
            out, self.buffer = self.buffer[:size], self.buffer[size:]
        return out

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class TableClient:
    """Tables store row-based data with a well-defined column schema. Such data may come from spreadsheets
    or other tabular data stores.

    Each table is created with a well-defined column schema, which can include primitive columns such
    as numbers and strings, as well as nested struct columns and arrays.
    A table's schema can be extended but existing columns cannot generally be modified or removed.

    Table data as well as it's column schema itself are version controlled. Each change to the contained
    rows is tracked as a data version.
    Tables can be queried at any historic version unless some destructive change has been made, such as
    hard deletion of historic data for compliance reasons.

    Because tables are versioned, each row is immutable. This means that it is not possible to change individual
    values in a row. Instead, the data load that added the row can be undone and the rows can be loaded
    again with the changed data.

    See the [Data Loads](#tag/Data-Loads) API endpoints on how to add and remove data to tables.
    """

    def __init__(self, client: HttpClient):
        self._client = client.with_prefix("/table")

    def query(
        self,
        query: str,
        *,
        project_id: int | None,
        version_id: int | None = None,
    ) -> Generator[pa.RecordBatch, None, None]:
        """Query data with the given SQL query and retrieve the result as arrow record batches.

        Args:
            query: The SQL query the execute.
            project_id: project_id to filter the source data by. Can be set to none to query data
                access all projects the authenticated user has access to.
            version_id: Data version ID at which to execute the query.

        Example:
            ```
            result = client.query("SELECT * FROM source_table")
            table = pyarrow.Table.from_batches(result)

            # Write as CSV. Does only work with primitive columns.
            pyarrow.csv.write_csv(t, "result.csv")

            # Write as Parquet. Does work with structs and array columns as well.
            pyarrow.parquet.write_table(t, "result.parquet")
            ```
        """
        request = QueryDataRequest(
            query=query,
            project_id=project_id,
            version_id=version_id,
        )
        with self._client.http_client.stream(
            "POST", self._client.url(":query"), json=request.model_dump(by_alias=True)
        ) as response:
            if not response.is_success:
                response.read()  # Required in streaming request to the error handling can access content.
                self._client.handle_error_response(response)

            ct = response.headers.get("content-type")
            if ct is None:
                raise ValueError("content type header missing")
            ct = ct.split(";")[0].strip()

            if ct != "application/vnd.apache.arrow.stream":
                raise ValueError(f"unexpected content type {ct}")
            yield from pa.ipc.open_stream(_IteratorReader(response.iter_bytes()))

    def get_by_id(self, table_id: int) -> TableResponse:
        """Retrieve the table by its ID.
        The ID is an opaque identifier for the table returned at table creation time and is distinct from the table reference.
        """
        return self._client.get(":getById", TableResponse, dict(id=table_id))

    def get(self, reference: str) -> TableResponse:
        """Get a table by its reference."""
        return self._client.get(":get", TableResponse, dict(reference=reference))

    def list(self) -> Generator[TableResponse, None, None]:
        """List all current base tables and views in the workspace."""
        opts = ListOptions()
        while True:
            resp = self._client.get(":list", ListTableResponse, params=opts.model_dump())
            yield from resp.items
            if resp.cursor is None:
                return
            opts.cursor = resp.cursor

    def list_versions(self, table_id: int) -> Generator[TableResponse, None, None]:
        """List all versions of a specific table."""
        opts = ListOptions()
        while True:
            resp = self._client.get(
                ":listVersions",
                ListTableResponse,
                params=dict(**opts.model_dump(), id=table_id),
            )
            yield from resp.items
            if resp.cursor is None:
                return
            opts.cursor = resp.cursor

    def list_archived(self, reference: str) -> Generator[TableResponse, None, None]:
        """List all archived tables that match the given reference."""
        opts = ListOptions()
        while True:
            resp = self._client.get(
                ":listArchived",
                ListTableResponse,
                params=dict(**opts.model_dump(), reference=reference),
            )
            yield from resp.items
            if resp.cursor is None:
                return
            opts.cursor = resp.cursor

    def create(self, table: ViewTableCreate | BaseTableCreate) -> TableResponse:
        """Create a new table or view."""
        return self._client.post(":create", TableResponse, json=table)

    def update(self, reference: str, table: ViewTableUpdate | BaseTableUpdate) -> TableResponse:
        """Update a table or view.

        Tables can only be updated in forward-compatible ways.
        New columns can be added but not be removed.
        Existing columns can be turned from non-nullable to nullable but not the other way around.
        Existing columns cannot be renamed or changed in type.

        Views can be updated arbitrarily.
        """
        return self._client.post(":update", TableResponse, params=dict(reference=reference), json=table)

    def archive(self, reference: str, archive: TableArchive | None = None) -> TableResponse:
        """Archive a table or view.

        Archiving makes the table or view inaccessible for regular operations
        but preserves it for historical purposes.
        """
        if not archive:
            archive = TableArchive()
        return self._client.post(":archive", TableResponse, params=dict(reference=reference), json=archive)

    def unarchive(self, table_id: int) -> TableResponse:
        """Unarchive a previously archived table or view.

        This makes the table or view available for regular operations again.
        If an active table with the same reference already exists, the unarchive operation
        will fail with an HTTP 409 `Conflict error`.
        """
        return self._client.post(":unarchive", TableResponse, params=dict(id=table_id))

    def rename(self, reference: str, rename: TableRename) -> TableResponse:
        """Rename the table."""
        return self._client.post(":rename", TableResponse, params=dict(reference=reference), json=rename)


class DataLoadClient:
    """The data API provides functionality for managing and interacting with data in a workspace.

    Data comes in the form of *tables* and *artifacts*.
    *Tables* are used to store row-based data, such as measurements, sequences, and other structured data.
    *Artifacts* are used to store the results of tasks, such as machine learning models, predictions, and other outputs.

    While tables can be created and uploaded by users, artifacts are only produced by tasks and cannot be uploaded directly.

    Artifacts and table rows (via their data load source) have a context specified. This context refers to the
    origin of the data and is used in other parts of the API to determine data visibility, e.g. to
    isolate data between projects.

    See the [Artifacts](#tag/Artifacts) API endpoints for more details on artifact data.

    See the [Tables](#tag/Tables) API endpoints for more details on tabular data.
    """

    def __init__(self, client: HttpClient):
        self._client = client.with_prefix("/load")

    def list(
        self, project_id: int | None = None, round_id: int | None = None
    ) -> Generator[DataLoadResponse, None, None]:
        """List all data loads in the workspace."""
        params = {}
        if project_id is not None:
            params["project_id"] = project_id
        if round_id is not None:
            params["round_id"] = round_id

        opts = ListOptions()
        while True:
            resp = self._client.get(":list", ListDataLoadResponse, params={**params, **opts.model_dump()})
            yield from resp.items
            if resp.cursor is None:
                return
            opts.cursor = resp.cursor

    def get(self, load_id: int) -> DataLoadResponse:
        """Get a data load by ID."""
        return self._client.get(":get", DataLoadResponse, dict(id=load_id))

    def create(self, load: DataLoadCreate) -> DataLoadResponse:
        """Create a new data load.
        After creation, the data load will be empty. Files can subsequently be added using the `uploadFile` endpoint.
        """
        return self._client.post(":create", DataLoadResponse, json=load)

    def upload_file(
        self,
        load_id: int,
        file: Path,
        *,
        description: str | None = None,
        filepath: str | Path | None = None,
        table_reference: str | None = None,
        source_file_id: int | None = None,
    ) -> FileUploadResponse:
        """Upload a file to the cloud storage bucket and add upload information to load."""
        data = {}
        if description is not None:
            data["description"] = description
        if filepath is not None:
            data["filepath"] = str(filepath)
        if table_reference is not None:
            data["table_reference"] = table_reference
        if source_file_id is not None:
            data["source_file_id"] = source_file_id

        with file.open("rb") as f:
            return self._client.post(
                ":uploadFile",
                FileUploadResponse,
                params=dict(id=load_id),
                files=dict(file=(file.name, f, "application/octet-stream")),
                data=data,
            )

    def download_file(self, file_id: str, buffer: BinaryIO):
        """Download a previously uploaded file from a load."""
        bytes_stream: Iterator[bytes] = self._client.get(":downloadFile", Iterator[bytes], dict(file_id=file_id))
        buffer.writelines(bytes_stream)

    def finalize(self, load_id: int) -> DataLoadResponse:
        """Finalize the data load and start ingesting rows from the uploaded files.

        The ingestion process will run in the background. The progress can be tracked by polling the data load
        and waiting for the status to switch from `LOADING` to `COMPLETED` or `FAILED`.
        """
        return self._client.post(":finalize", DataLoadResponse, params=dict(id=load_id))

    def delete(self, load_id: int) -> DataLoadResponse:
        """Delete the data load if it has not been completed yet.

        The deletion will happen asynchronously afterwards and is irreversible.
        """
        return self._client.post(":delete", DataLoadResponse, params=dict(id=load_id))

    def undo(self, load_id: int) -> DataLoadResponse:
        """Append an entry to the changelog undoing the load.

        The rows from this data load will no longer show at subsequent data versions.
        Previously undone loads can be redone to restore their data.
        """
        return self._client.post(":undo", DataLoadResponse, params=dict(id=load_id))

    def redo(self, load_id: int) -> DataLoadResponse:
        """Append an entry to the changelog redoing the load. The rows from this data load will be visible again at subsequent data versions."""
        return self._client.post(":redo", DataLoadResponse, params=dict(id=load_id))

    def add_table(self, load_id: int, request: AddTableRequest) -> DataLoadResponse:
        """Register a new table in the data load configuration."""
        return self._client.post(":addTable", DataLoadResponse, params=dict(id=load_id), json=request)


class TaskClient:
    """Create endpoints for all supported task types.

    When creating a task, a context must be specified. The context determines data visibility: data can either be visible
    in the entire project, or only in a specific round within a project.
    Artifact and table data also has a context and a task can only access data with a compatible context. A data context
    is compatible when it a workspace context or, if it is a project or round context, belongs to the same project.

    For example, suppose project `A` has rounds `abc` and `def` and project `B` has round `xyz`.
    A task with context `project=A` or `round=abc` can only see data with context `project=A` or `round=abc` or `round=def`.
    A task with context `project=B` can only see data with context `project=B` or `round=xyz`.

    Any data a task creates will be saved with the same context as the task.

    Tasks always use the latest version of the data in the workspace, unless a specific version is provided.
    """

    def __init__(self, client: HttpClient):
        self._client = client.with_prefix("/task")

    def list(self, project_id: int | None = None, round_id: int | None = None) -> Generator[TaskResponse, None, None]:
        """List all tasks in a workspace, project or round.

        To list all tasks in a workspace, leave `project_id` and `round_id` unset.
        To list all tasks in a project, specify `project_id` and leave `round_id` unset.
        To list all tasks in a round, specify `round_id` and leave `project_id` unset.

        Only one of `project_id` or `round_id` must be specified at once.
        """
        params = {}
        if project_id is not None:
            params["project_id"] = project_id
        if round_id is not None:
            params["round_id"] = round_id

        opts = ListOptions()
        while True:
            resp = self._client.get(":list", ListTaskResponse, params={**params, **opts.model_dump()})
            yield from resp.items
            if resp.cursor is None:
                return
            opts.cursor = resp.cursor

    def get(self, task_id: int) -> TaskResponse:
        """Obtain the current state and the result of a task.

        This endpoint can be called periodically to monitor the status of a task and to detect
        when it finished executing.

        If the `state` field of the response is `COMPLETED`, then the `result` field contains the
        task result (often containing IDs of artifacts and tables created by the task).

        If the `state` field of the response is `FAILED`, then the `error` field contains
        a description explaining why the task has failed.

        The states `COMPLETED`, `FAILED` and `CANCELLED` are terminal states. When the task is in
        any other state, it will eventually transition into one of these three.
        """
        return self._client.get(":get", TaskResponse, dict(id=task_id))

    def get_by_name(
        self,
        name: str,
        project_id: int | None = None,
        round_id: int | None = None,
    ) -> TaskResponse:
        """Get a task by unique name within their workspace, project, or round context.

        If a task with name "TaskA" exists in project with ID 123, the `project_id` must be specified.
        It is not sufficient to leave it unset as tasks with the same name may exist at the workspace or
        round context or in a different project.
        """
        tasks = self.list(project_id=project_id, round_id=round_id)
        # When listing for workspace or project we still need to filter it down since they'll
        # return tasks from their sub-contexts.
        if round_id is not None:
            scope_desc = f"round {round_id}"
            tasks = [task for task in tasks if task.context == ContextRound(round_id=round_id)]
        elif project_id is not None:
            scope_desc = f"project {project_id}"
            tasks = [task for task in tasks if task.context == ContextProject(project_id=project_id)]
        else:
            scope_desc = "workspace"
            tasks = [task for task in tasks if task.context == ContextWorkspace()]

        for task in tasks:
            if task.name == name:
                return task
        raise ValueError(f"Task with name '{name}' not found in {scope_desc}")

    def create(self, task: TaskCreate) -> TaskResponse:
        return self._client.post(f"/{task.parameters.task_type}:create", TaskResponse, json=task)

    def create_or_get(self, task: TaskCreate, skip_parameter_check: bool = False) -> TaskResponse:
        """Create a task or get existing one with matching (name or idempotency key) and parameters.

        The task must have a name or an idempotency key set. If the parameter check is not skipped the existing task must have
        the same parameters as `task`. If an explicit `data_version_id` is specified, it must also match.

        Raises ValueError if parameters don't match or if name is None.
        """

        error_prefix: str = ""

        # If you're using idempotency_key, the :create endpoint is already an upsert.
        if task.idempotency_key is not None:
            error_prefix = f"Task with idempotency key '{task.idempotency_key}'"
            existing_task = self.create(task)
        elif name := task.name:
            error_prefix = f"Task with name '{name}'"
            try:
                if isinstance(task.context, ContextRound):
                    existing_task = self.get_by_name(name, round_id=task.context.round_id)
                elif isinstance(task.context, ContextProject):
                    existing_task = self.get_by_name(name, project_id=task.context.project_id)
                else:
                    existing_task = self.get_by_name(name)
            except ValueError:
                existing_task = None
        else:
            raise ValueError("Task must have `name` or `idempotency_key` set")

        if existing_task is None:
            return self.create(task)

        if skip_parameter_check:
            return existing_task

        if existing_task.parameters != task.parameters:
            existing_params = existing_task.parameters.model_dump_json(indent=2, by_alias=True)
            new_params = task.parameters.model_dump_json(indent=2, by_alias=True)
            diff = "\n".join(
                difflib.unified_diff(
                    existing_params.splitlines(),
                    new_params.splitlines(),
                    fromfile="existing_parameters",
                    tofile="new_parameters",
                    lineterm="",
                )
            )
            raise ValueError(f"{error_prefix} exists but has different parameters:\n{diff}")

        if task.data_version_id is not None and existing_task.data_version_id != task.data_version_id:
            raise ValueError(
                f"{error_prefix} exists but has different data_version_id (existing: {existing_task.data_version_id}, new: {task.data_version_id})"
            )

        return existing_task

    def cancel(self, task_id: int) -> TaskResponse:
        """Cancel the task if possible.

        Only tasks that have not yet terminated may be canceled. There is no guarantee that cancellation succeeds
        before the task successfully completes.
        """
        return self._client.post(":cancel", TaskResponse, dict(id=task_id))

    def recover(self, task_id: int) -> TaskResponse:
        """Recover a task that has failed or been cancelled by putting it into state EXECUTING.

        Note that this is best-effort and is not guaranteed to fix any issues in the underlying
        backend execution, and may also cause unexpected side-effects such as duplicate data loads.
        Main use-case is tasks that failed due to intermittent issues in the backend which have been resolved.
        """
        return self._client.post(":recover", TaskResponse, dict(id=task_id))

    def archive(self, task_id: int) -> TaskResponse:
        """Archive the task.

        This will make the task disappear from regular task listings. It will not remove data that has been
        produced by the task. If data removal is required, use the `data_load_id` for the task's results
        and call the `/data/load:undo` endpoint separately.
        """
        return self._client.post(":archive", TaskResponse, dict(id=task_id))

    def unarchive(self, task_id: int) -> TaskResponse:
        """Unarchive the task."""
        return self._client.post(":unarchive", TaskResponse, dict(id=task_id))

    def wait(self, task_id: int, timeout: float = 60.0) -> TaskResponse:
        start = time.monotonic()

        while time.monotonic() - start < timeout:
            try:
                task = self.get(task_id)
                if task.state == TaskState.FAILED:
                    raise ValueError(f"Task {task_id} failed: {task.errors}")
                if task.state == TaskState.CANCELLED:
                    raise ValueError(f"Task {task_id} was cancelled")
                if task.state == TaskState.COMPLETED:
                    break
            except httpx.TransportError:
                logger.warning("Transport error while waiting for task, retrying...")
            except ClientError as e:
                if 500 <= e.status_code < 600:
                    logger.warning("Internal server error while waiting for task, retrying...")
                else:
                    raise

            time.sleep(3)
        print()
        return self.get(task_id)



class ClientError(Exception):
    def __init__(self, status_code: int, message: str, errors: list[str]):
        self.status_code = status_code
        self.message = message

        message = f"API Error {status_code}: {message}"
        if len(errors) > 0:
            error_list = "\n".join(f"\t{error}" for error in errors)
            message = f"{message}\n{error_list}"

        super().__init__(message)



import json
import time
from pathlib import Path

from cradle.sdk.types.data import FileUploadResponse
from cradle.sdk.types.task import TaskResponse

from .client import DataLoadClient, TaskClient


def upload_blob_files(
    client: DataLoadClient, load_id: int, files: list[Path], table_reference: str | None = None
) -> list[FileUploadResponse]:
    responses = []
    for f in files:
        response = client.upload_file(load_id=load_id, file=f, table_reference=table_reference)
        print(f"Uploaded {f}, id={response.id}")
        responses.append(response)
    return responses


def wait_for_task(client: TaskClient, task_id: int, timeout: float = 60) -> TaskResponse:
    start = time.monotonic()
    while time.monotonic() - start < timeout:
        task = client.get(task_id)
        print(f"\r{task.state} @ {task.updated_at}", end="", flush=True)
        if task.state == "COMPLETED":
            if task.result is None:
                raise ValueError("Task result is None")
            print()
            print(json.dumps(task.result.model_dump(mode="json"), indent=2))
            return task
        if task.state == "FAILED":
            print()
            print(task.errors)
            return task
        if task.state == "CANCELLED":
            return task
        time.sleep(1)
    return client.get(task_id)



from __future__ import annotations

from datetime import datetime
from typing import Annotated, Literal, TypeAlias

from pydantic import BaseModel, Field


class AbstractResourceResponse(BaseModel):
    id: int = Field(description="The unique identifier of the resource")
    created_at: datetime = Field(description="The time at which the resource was created")
    created_by: str = Field(description="The ID of the user who created the resource")


class ArchivableResourceMixin(BaseModel):
    archived_at: datetime | None = Field(default=None, description="The time at which the resource was archived")
    archived_by: str | None = Field(default=None, description="The ID of the user who archived the resource")


class BaseListResponse(BaseModel):
    items: list[AbstractResourceResponse] = Field(description="The listed resources.")
    cursor: str | None = Field(
        description="Cursor with which the next page of entries can be retrieved. If no cursor is returned, no further pages exist."
    )


class ContextProject(BaseModel):
    kind: Literal["PROJECT"] = Field(default="PROJECT")
    project_id: int


class ContextRound(BaseModel):
    kind: Literal["ROUND"] = Field(default="ROUND")
    round_id: int


class ContextWorkspace(BaseModel):
    kind: Literal["WORKSPACE"] = Field(default="WORKSPACE")


class ErrorResponseMixin(BaseModel):
    """Mixin for resources that can return errors."""

    errors: list[str] = Field(default_factory=list, description="List of error messages associated with the resource.")


class ListOptions(BaseModel):
    limit: int = Field(default=100, description="Maximum number of items to return")
    cursor: str | None = Field(default=None, description="Continuation token returned by the previous page's response")
    order: Literal["ASC", "DESC"] = Field(
        default="ASC", description="Whether to order the result ascending or descending"
    )


class ImmutableResourceResponse(AbstractResourceResponse):
    pass


class ResourceResponse(AbstractResourceResponse):
    updated_at: datetime = Field(description="The time at which the resource was last updated")
    updated_by: str = Field(description="The ID of the user who last updated the resource")


Context: TypeAlias = Annotated[ContextWorkspace | ContextProject | ContextRound, Field(discriminator="kind")]



from __future__ import annotations

from abc import ABC
from enum import StrEnum
from typing import Literal

from pydantic import BaseModel, Field


class TaskState(StrEnum):
    INIT = "INIT"
    PREPARING_INPUTS = "PREPARING_INPUTS"
    LAUNCHING = "LAUNCHING"
    EXECUTING = "EXECUTING"
    LOADING_RESULTS = "LOADING_RESULTS"
    CANCELLING = "CANCELLING"
    RECOVERING = "RECOVERING"
    FAILED = "FAILED"
    COMPLETED = "COMPLETED"
    CANCELLED = "CANCELLED"


class ArtifactParam(BaseModel):
    artifact_id: int


class ParametersBase(BaseModel, ABC):
    """Class that marks top-level task inputs. Used for static checks."""


class ResultBase(BaseModel, ABC):
    """Class that marks top-level task results. Used for static checks."""


class TableOutputBase(BaseModel):
    table: str = Field(description="The reference to the table the result was written to.")


class TableSourceQuery(BaseModel):
    kind: Literal["QUERY"] = Field(default="QUERY")
    query: str


class TableSourceTable(BaseModel):
    kind: Literal["TABLE"] = Field(default="TABLE")
    table: str


class TableSourceTaskResult(BaseModel):
    kind: Literal["TASK_RESULT"] = Field(default="TASK_RESULT")
    table: str
    task_id: int | list[int]


class TableResult(TableOutputBase):
    pass
